# -*- coding: utf-8 -*-
# Copyright 2024, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of eoreader project
#     https://github.com/sertit/eoreader
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Sentinel-2 products """
import difflib
import json
import logging
import os
import re
import tempfile
import zipfile
from datetime import datetime
from enum import unique
from typing import Union

import geopandas as gpd
import numpy as np
import pandas as pd
import planetary_computer
import rasterio
import xarray as xr
from affine import Affine
from lxml import etree
from pystac import Item
from rasterio import errors, features, transform
from rasterio.crs import CRS
from rasterio.enums import Resampling
from sertit import AnyPath, files, geometry, path, rasters, vectors
from sertit.misc import ListEnum
from sertit.types import AnyPathStrType, AnyPathType
from shapely.geometry import box
from stac_asset import S3Client

from eoreader import DATETIME_FMT, EOREADER_NAME, cache, utils
from eoreader.bands import (
    ALL_CLOUDS,
    BLUE,
    CA,
    CIRRUS,
    CLOUDS,
    EOREADER_STAC_MAP,
    GREEN,
    NARROW_NIR,
    NIR,
    RAW_CLOUDS,
    RED,
    SHADOWS,
    SWIR_1,
    SWIR_2,
    SWIR_CIRRUS,
    VRE_1,
    VRE_2,
    VRE_3,
    WV,
    BandNames,
    SpectralBand,
    to_str,
)
from eoreader.exceptions import InvalidProductError, InvalidTypeError
from eoreader.products import OpticalProduct, StacProduct
from eoreader.products.optical.optical_product import RawUnits
from eoreader.products.product import OrbitDirection
from eoreader.stac import CENTER_WV, FWHM, GSD, ID, NAME
from eoreader.utils import simplify

LOGGER = logging.getLogger(EOREADER_NAME)


@unique
class S2ProductType(ListEnum):
    """Sentinel-2 products types (L1C or L2A)"""

    L1C = "MSIL1C"
    """Level-1C: https://sentinel.esa.int/web/sentinel/user-guides/sentinel-2-msi/product-types/level-1c"""

    L2A = "MSIL2A"
    """Level-2A: https://sentinel.esa.int/web/sentinel/user-guides/sentinel-2-msi/product-types/level-2a"""


@unique
class S2GmlMasks(ListEnum):
    """Sentinel-2 GML masks (processing baseline < 4.0)"""

    FOOTPRINT = "DETFOO"
    CLOUDS = "CLOUDS"
    DEFECT = "DEFECT"
    NODATA = "NODATA"
    SATURATION = "SATURA"
    QUALITY = "TECQUA"

    # L2A (jp2)
    CLDPRB = "CLDPRB"
    SNWPRB = "SNWPRB"


@unique
class S2Jp2Masks(ListEnum):
    """Sentinel-2 jp2 masks (processing baseline > 4.0)"""

    # Both L1C and L2A
    FOOTPRINT = "DETFOO"
    CLOUDS = "CLASSI"
    QUALITY = "QUALIT"  # Regroups TECQUA, DEFECT, NODATA, SATURA

    # L2A
    CLDPRB = "CLDPRB"
    SNWPRB = "SNWPRB"


BAND_DIR_NAMES = {
    S2ProductType.L1C: ".",
    S2ProductType.L2A: {
        "01": ["R60m"],
        "02": ["R10m", "R20m", "R60m"],
        "03": ["R10m", "R20m", "R60m"],
        "04": ["R10m", "R20m", "R60m"],
        "05": ["R20m", "R60m"],
        "06": ["R20m", "R60m"],
        "07": ["R20m", "R60m"],
        "08": ["R10m"],
        "8A": ["R20m", "R60m"],
        "09": ["R60m"],
        "11": ["R20m", "R60m"],
        "12": ["R20m", "R60m"],
    },
}


class S2Product(OpticalProduct):
    """
    Class of Sentinel-2 Products

    You can use directly the .zip file
    """

    def __init__(
        self,
        product_path: AnyPathStrType,
        archive_path: AnyPathStrType = None,
        output_path: AnyPathStrType = None,
        remove_tmp: bool = False,
        **kwargs,
    ) -> None:
        # Processing baseline < 02.07: images not georeferenced (L2Ap and after)

        # Is this product comes from a processing baseline less than 4.0
        # The processing baseline 4.0 introduces format changes:
        # - masks are given as GeoTIFFs instead of GML files
        # - an offset is added to keep the zero as no-data value
        # See here for more information
        # https://sentinels.copernicus.eu/web/sentinel/-/copernicus-sentinel-2-major-products-upgrade-upcoming
        self._processing_baseline = None
        self.raw_no_data = 0
        self.no_data_val = {}

        # L2Ap
        self._is_l2ap = False

        # S2 Sinergise
        self._is_sinergise = kwargs.pop("is_sinergise", False)

        # Initialization from the super class
        super().__init__(product_path, archive_path, output_path, remove_tmp, **kwargs)

        try:
            self.read_mtd()
        except InvalidProductError:
            LOGGER.warning(
                f"Corrupted metadata for {self.path}. "
                f"Trying to process this product in degraded mode. "
                f"Every process needing something from the metadata won't be able to be computed (i.e. HILLSHADE)"
            )

    def _pre_init(self, **kwargs) -> None:
        """
        Function used to pre_init the products
        (setting needs_extraction and so on)
        """
        self._has_cloud_cover = True
        self.needs_extraction = False
        # Use filename for SAFE names, not for others
        # S2A_MSIL1C_20191215T110441_N0208_R094_T30TXP_20191215T114155.SAFE has 65 characters
        self._use_filename = len(self.filename) > 50
        self._raw_units = RawUnits.REFL

        # We need to set the constellation asap for this product (to manage correctly the name of broken products)
        self.constellation = self._get_constellation()

        # Pre init done by the super class
        super()._pre_init(**kwargs)

    def _post_init(self, **kwargs) -> None:
        """
        Function used to post_init the products
        (setting sensor type, band names and so on)
        """
        self.tile_name = self._get_tile_name()

        # Get processing baseline: N0213 -> 02.13
        pr_baseline = float(self.split_name[3][1:]) / 100
        self._processing_baseline = pr_baseline

        # Post init done by the super class
        super()._post_init(**kwargs)

    def _set_pixel_size(self) -> None:
        """
        Set product default pixel size (in meters)
        """
        # S2: use 10m resolution, even if we have 60m and 20m resolution
        # In the future maybe use one resolution per band ?
        self.pixel_size = 10.0

    def _get_tile_name(self) -> str:
        """
        Retrieve tile name

        Returns:
            str: Tile name
        """
        return self.split_name[-2]

    def _set_product_type(self) -> None:
        """Set products type"""
        self.product_type = S2ProductType.from_value(self.split_name[1])

    def _set_instrument(self) -> None:
        """
        Set instrument

        Sentinel-2: https://sentinels.copernicus.eu/web/sentinel/missions/sentinel-2/instrument-payload/
        """
        self.instrument = "MSI"

    def _map_bands(self) -> None:
        """
        Map bands
        """
        l2a_bands = {
            CA: SpectralBand(
                eoreader_name=CA,
                **{NAME: "B01", ID: "01", GSD: 60, CENTER_WV: 442, FWHM: 21},
            ),
            BLUE: SpectralBand(
                eoreader_name=BLUE,
                **{NAME: "B02", ID: "02", GSD: 10, CENTER_WV: 492, FWHM: 66},
            ),
            GREEN: SpectralBand(
                eoreader_name=GREEN,
                **{NAME: "B03", ID: "03", GSD: 10, CENTER_WV: 560, FWHM: 36},
            ),
            RED: SpectralBand(
                eoreader_name=RED,
                **{NAME: "B04", ID: "04", GSD: 10, CENTER_WV: 665, FWHM: 31},
            ),
            VRE_1: SpectralBand(
                eoreader_name=VRE_1,
                **{NAME: "B05", ID: "05", GSD: 20, CENTER_WV: 704, FWHM: 15},
            ),
            VRE_2: SpectralBand(
                eoreader_name=VRE_2,
                **{NAME: "B06", ID: "06", GSD: 20, CENTER_WV: 740, FWHM: 15},
            ),
            VRE_3: SpectralBand(
                eoreader_name=VRE_3,
                **{NAME: "B07", ID: "07", GSD: 20, CENTER_WV: 781, FWHM: 20},
            ),
            NIR: SpectralBand(
                eoreader_name=NIR,
                **{NAME: "B08", ID: "08", GSD: 10, CENTER_WV: 833, FWHM: 106},
            ),
            NARROW_NIR: SpectralBand(
                eoreader_name=NARROW_NIR,
                **{NAME: "B8A", ID: "8A", GSD: 20, CENTER_WV: 864, FWHM: 21},
            ),
            WV: SpectralBand(
                eoreader_name=WV,
                **{NAME: "B09", ID: "09", GSD: 60, CENTER_WV: 944, FWHM: 20},
            ),
            SWIR_1: SpectralBand(
                eoreader_name=SWIR_1,
                **{NAME: "B11", ID: "11", GSD: 20, CENTER_WV: 1612, FWHM: 92},
            ),
            SWIR_2: SpectralBand(
                eoreader_name=SWIR_2,
                **{NAME: "B12", ID: "12", GSD: 20, CENTER_WV: 2190, FWHM: 180},
            ),
        }

        if self.product_type == S2ProductType.L2A:
            self.bands.map_bands(l2a_bands)
        elif self.product_type == S2ProductType.L1C:
            self.bands.map_bands(
                {
                    **l2a_bands,
                    SWIR_CIRRUS: SpectralBand(
                        eoreader_name=SWIR_CIRRUS,
                        **{NAME: "B10", ID: "10", GSD: 60, CENTER_WV: 1380, FWHM: 30},
                    ),
                }
            )
        else:
            raise InvalidProductError(f"Invalid Sentinel-2 name: {self.filename}")

    @cache
    def crs(self) -> CRS:
        """
        Get UTM projection of the tile

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.crs()
            CRS.from_epsg(32630)

        Returns:
            rasterio.crs.CRS: CRS object
        """
        if self._processing_baseline < 2.07:
            try:
                root, ns = self.read_mtd()
                crs = CRS.from_string(root.findtext(".//HORIZONTAL_CS_CODE"))
            except InvalidProductError:
                # Manage broken XML
                utm_nb = self.tile_name[1:3]
                utm_letter = self.tile_name[3]
                utm_hemisphere = 6 if utm_letter > "N" else 7
                crs = CRS.from_string(f"epsg:32{utm_hemisphere}{utm_nb}")
        else:
            crs = super().crs()

        return crs

    @cache
    def extent(self) -> gpd.GeoDataFrame:
        """
        Get UTM extent of the tile

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.extent()
                                                        geometry
            0  POLYGON ((309780.000 4390200.000, 309780.000 4...

        Returns:
            gpd.GeoDataFrame: Extent in UTM
        """
        if self._processing_baseline < 2.07:
            tf, width, height, crs = self.default_transform()
            bounds = transform.array_bounds(height, width, tf)
            return gpd.GeoDataFrame(geometry=[box(*bounds)], crs=crs)
        else:
            return super().extent()

    @cache
    @simplify
    def footprint(self) -> gpd.GeoDataFrame:
        """
        Get UTM footprint in UTM of the products (without nodata, *in french == emprise utile*)

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.footprint()
               index                                           geometry
            0      0  POLYGON ((199980.000 4500000.000, 199980.000 4...

        Returns:
            gpd.GeoDataFrame: Footprint as a GeoDataFrame
        """
        def_band = self.bands[self.get_default_band()].id
        if self._processing_baseline < 4.0:
            det_footprint = self._open_mask_lt_4_0(S2GmlMasks.FOOTPRINT, def_band)
            footprint_gs = det_footprint.dissolve().convex_hull
            footprint = gpd.GeoDataFrame(
                geometry=footprint_gs.geometry, crs=footprint_gs.crs
            )

            # Manage broken GML
            if all(footprint.is_empty):
                try:
                    LOGGER.warning(
                        "Invalid DETFOO mask. Trying to vectorize nodata from GREEN band. Your product may be broken and the results may be inaccurate!"
                    )

                    footprint = rasters.vectorize(
                        det_footprint, values=0, keep_values=False, dissolve=True
                    )

                    footprint = geometry.get_wider_exterior(footprint).to_crs(
                        self.crs()
                    )
                except Exception:
                    LOGGER.error(
                        "Impossible to return the footprint. Returning the extent instead."
                    )
                    footprint = self.extent()

        else:
            det_footprint = self._open_mask_gt_4_0(S2Jp2Masks.FOOTPRINT, def_band)
            footprint = rasters.vectorize(
                det_footprint, values=0, keep_values=False, dissolve=True
            )

            # Keep only the convex hull
            footprint.geometry = footprint.geometry.convex_hull

        return footprint

    def get_datetime(self, as_datetime: bool = False) -> Union[str, datetime]:
        """
        Get the product's acquisition datetime, with format :code:`YYYYMMDDTHHMMSS` <-> :code:`%Y%m%dT%H%M%S`

        .. WARNING::
            Sentinel-2 datetime is the datatake sensing time, not the granule sensing time !
            (the one displayed in the product's name)

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.get_datetime(as_datetime=True)
            datetime.datetime(2020, 8, 24, 11, 6, 31)
            >>> prod.get_datetime(as_datetime=False)
            '20200824T110631'

        Args:
            as_datetime (bool): Return the date as a datetime.datetime. If false, returns a string.

        Returns:
             Union[str, datetime.datetime]: Its acquisition datetime
        """
        if self.datetime is None:
            # Sentinel-2 datetime (in the filename) is the datatake sensing time, not the granule sensing time !
            sensing_time = self.split_name[2]

            # Convert to datetime
            date = datetime.strptime(sensing_time, "%Y%m%dT%H%M%S")
        else:
            date = self.datetime

        if not as_datetime:
            date = date.strftime(DATETIME_FMT)

        return date

    def _get_name_constellation_specific(self) -> str:
        """
        Set product real name from metadata

        Returns:
            str: True name of the product (from metadata)
        """
        try:
            # Get MTD XML file
            root, _ = self.read_datatake_mtd()

            # Open identifier
            name = root.findtext(".//PRODUCT_URI")
            if not name:
                # Manage L2Ap products
                name = root.findtext(".//PRODUCT_URI_2A")
                if not name:
                    raise InvalidProductError("PRODUCT_URI not found in metadata!")

            name = path.get_filename(name)
        except InvalidProductError:
            try:
                tile_info = files.read_json(
                    next(self.path.glob("**/tileInfo.json")), print_file=False
                )
                name = tile_info["productName"]
            except json.JSONDecodeError:
                raise InvalidProductError(
                    f"Corrupted metadata and bad filename for {self.path}! "
                    f"Impossible to process this product."
                )

        return name

    def _get_qi_folder(self):
        """"""
        if self._is_sinergise:
            mask_folder = "qi"
        elif self.is_archived:
            mask_folder = ".*GRANULE.*QI_DATA"
        else:
            mask_folder = "**/*GRANULE/*/QI_DATA"

        return mask_folder

    def _get_image_folder(self):
        """"""
        if self._is_sinergise:
            img_folder = "."
        elif self.is_archived:
            img_folder = ".*GRANULE.*IMG_DATA"
        else:
            img_folder = "**/*GRANULE/*/IMG_DATA"

        return img_folder

    def _get_res_band_folder(self, band_list: list, pixel_size: float = None) -> dict:
        """
        Return the folder containing the bands of a proper S2 products.
        (IMG_DATA for L1C, IMG_DATA/Rx0m for L2A)

        Args:
            band_list (list): Wanted bands (listed as 01, 02...)
            pixel_size (float): Band resolution for Sentinel-2 products {R10m, R20m, R60m}.
                                The wanted bands will be chosen in this proper folder.

        Returns:
            dict: Dictionary containing the folder path for each queried band
        """
        if pixel_size is not None:
            if isinstance(pixel_size, (list, tuple)):
                pixel_size = pixel_size[0]

        # Open the band directory names
        s2_bands_folder = {}

        # Manage L2A
        band_dir = BAND_DIR_NAMES[self.product_type]
        for band in band_list:
            band_id = self.bands[band].id
            if band_id is None:
                raise InvalidProductError(
                    f"Non existing band ({band.name}) for S2-{self.product_type.name} products"
                )

            # If L2A products, we care about the resolution
            if self.product_type == S2ProductType.L2A:
                # If we got a true S2 resolution, open the corresponding band
                if pixel_size and f"R{int(pixel_size)}m" in band_dir[band_id]:
                    dir_name = f"R{int(pixel_size)}m"

                # Else open the first one, it will be resampled when the band will be read
                else:
                    dir_name = band_dir[band_id][0]
            # If L1C, we do not
            else:
                dir_name = band_dir

            if self.is_archived:
                # Open the zip file
                with zipfile.ZipFile(self.path, "r") as zip_ds:
                    # Get the band folder (use dirname is the first of the list is a band)
                    band_path = [
                        os.path.dirname(f.filename)
                        for f in zip_ds.filelist
                        if dir_name in f.filename
                    ][0]

                    # Workaround for a bug involving some bad archives
                    if band_path.startswith("/"):
                        band_path = band_path[1:]
                    s2_bands_folder[band] = band_path
            else:
                # Search for the name of the folder into the S2 products
                try:
                    s2_bands_folder[band] = next(
                        self.path.glob(f"{self._get_image_folder()}/{dir_name}")
                    )
                except IndexError:
                    s2_bands_folder[band] = self.path

        for band in band_list:
            if band not in s2_bands_folder:
                raise InvalidProductError(
                    f"Band folder for band {band.value} not found in {self.path}"
                )

        return s2_bands_folder

    def get_band_paths(
        self, band_list: list, pixel_size: float = None, **kwargs
    ) -> dict:
        """
        Return the paths of required bands.

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> from eoreader.bands import *
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.get_band_paths([GREEN, RED])
            {
                <SpectralBandNames.GREEN: 'GREEN'>: 'zip+file://S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip!/S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE/GRANULE/L1C_T30TTK_A027018_20200824T111345/IMG_DATA/T30TTK_20200824T110631_B03.jp2',
                <SpectralBandNames.RED: 'RED'>: 'zip+file://S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip!/S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE/GRANULE/L1C_T30TTK_A027018_20200824T111345/IMG_DATA/T30TTK_20200824T110631_B04.jp2'
            }

        Args:
            band_list (list): List of the wanted bands
            pixel_size (float): Band pixel size
            kwargs: Other arguments used to load bands

        Returns:
            dict: Dictionary containing the path of each queried band
        """
        band_folders = self._get_res_band_folder(band_list, pixel_size)
        band_paths = {}
        for band in band_list:
            # Get clean band path
            clean_band = self._get_clean_band_path(
                band, pixel_size=pixel_size, **kwargs
            )
            if clean_band.is_file():
                band_paths[band] = clean_band
            else:
                band_id = self.bands[band].id
                try:
                    if self.is_archived:
                        band_paths[band] = path.get_archived_rio_path(
                            self.path,
                            f".*{band_folders[band]}.*B{band_id}.*.jp2",
                        )
                    else:
                        band_paths[band] = path.get_file_in_dir(
                            band_folders[band],
                            f"B{band_id}",
                            extension="jp2",
                        )
                except (FileNotFoundError, IndexError) as ex:
                    raise InvalidProductError(
                        f"Non existing {band} ({band_id}) band for {self.path}"
                    ) from ex

        return band_paths

    def _read_band(
        self,
        band_path: AnyPathStrType,
        band: BandNames = None,
        pixel_size: Union[tuple, list, float] = None,
        size: Union[list, tuple] = None,
        **kwargs,
    ) -> xr.DataArray:
        """
        Read band from disk.

        .. WARNING::
            Invalid pixels are not managed here

        Args:
            band_path (AnyPathType): Band path
            band (BandNames): Band to read
            pixel_size (Union[tuple, list, float]): Size of the pixels of the wanted band, in dataset unit (X, Y)
            size (Union[tuple, list]): Size of the array (width, height). Not used if pixel_size is provided.
            kwargs: Other arguments used to load bands
        Returns:
            xr.DataArray: Band xarray

        """
        geocoded_path = band_path

        # For L2Ap
        try:
            if self._processing_baseline < 2.07 and str(band_path).endswith(".jp2"):
                # Get and write geocode data if not already existing
                with rasterio.open(str(band_path), "r") as ds:
                    if not ds.crs:
                        # Download path just in case
                        on_disk_path = (
                            self._get_band_folder(writable=True) / band_path.name
                        )
                        if not on_disk_path.is_file():
                            if path.is_cloud_path(band_path):
                                geocoded_path = band_path.download_to(
                                    self._get_band_folder(writable=True)
                                )
                            else:
                                geocoded_path = files.copy(
                                    band_path, self._get_band_folder(writable=True)
                                )
                        else:
                            geocoded_path = on_disk_path

                        # Get and write geocode data if not already existing
                        try:
                            with rasterio.open(str(geocoded_path), "r+") as out_ds:
                                tf, _, _, crs = self._l2ap_geocode_data(band_path)
                                out_ds.crs = crs
                                out_ds.transform = tf
                        except SystemError:
                            # Workaround for jp2 file that for a reason or another fails to be updated
                            # Maybe linked to https://github.com/rasterio/rasterio/issues/2528?
                            jp2_geocoded_path = geocoded_path
                            geocoded_path = jp2_geocoded_path.with_suffix(".tif")
                            with rasterio.open(str(jp2_geocoded_path), "r") as jp2_ds:
                                tif_meta = jp2_ds.meta
                                tif_meta["driver"] = "GTiff"
                                with rasterio.open(
                                    str(geocoded_path), "w", **tif_meta
                                ) as out_ds:
                                    out_ds.write(jp2_ds.read())
                                    tf, _, _, crs = self._l2ap_geocode_data(band_path)
                                    out_ds.crs = crs
                                    out_ds.transform = tf

        except errors.RasterioIOError as ex:
            if (
                str(band_path).endswith("jp2") or str(band_path).endswith("tif")
            ) and band_path.exists():
                raise InvalidProductError(f"Corrupted file: {band_path}") from ex
            else:
                raise ex

        # Read band
        return utils.read(
            geocoded_path,
            pixel_size=pixel_size,
            size=size,
            resampling=Resampling.bilinear,
            **kwargs,
        )

    def _to_reflectance(
        self,
        band_arr: xr.DataArray,
        band_path: AnyPathType,
        band: BandNames,
        **kwargs,
    ) -> xr.DataArray:
        """
        Converts band to reflectance

        Args:
            band_arr (xr.DataArray): Band array to convert
            band_path (AnyPathType): Band path
            band (BandNames): Band to read
            **kwargs: Other keywords

        Returns:
            xr.DataArray: Band in reflectance
        """
        # Only on raw files
        if str(band_path).endswith(".jp2") or (
            self._processing_baseline < 2.07
            and path.get_filename(band_path).startswith("T")
        ):
            try:
                # Get MTD XML file
                root, _ = self.read_datatake_mtd()

                # Get quantification value
                quantif_prefix = (
                    "BOA_" if self.product_type == S2ProductType.L2A else ""
                )
                try:
                    quantif_value = float(
                        root.findtext(f".//{quantif_prefix}QUANTIFICATION_VALUE")
                    )
                except TypeError:
                    raise InvalidProductError(
                        f"{quantif_prefix}QUANTIFICATION_VALUE not found in datatake metadata!"
                    )

                # Get offset
                offset_prefix = (
                    "BOA_" if self.product_type == S2ProductType.L2A else "RADIO_"
                )
                if self._processing_baseline < 4.0:
                    offset = 0.0
                else:
                    try:
                        if band == NARROW_NIR:
                            band_id = 8
                        else:
                            band_id = int(self.bands[band].id)
                        offset = float(
                            root.findtext(
                                f".//{offset_prefix}ADD_OFFSET[@band_id = '{band_id}']"
                            )
                        )
                    except TypeError:
                        raise InvalidProductError(
                            f"{offset_prefix}ADD_OFFSET not found in datatake metadata!"
                        )
            except InvalidProductError:
                # If not datatake file
                if self._processing_baseline < 4.0:
                    offset = 0.0
                else:
                    offset = -1000.0
                quantif_value = 10000.0

            # Compute the correct radiometry of the band
            band_arr = (band_arr + offset) / quantif_value

            self.no_data_val[band] = (self.raw_no_data + offset) / quantif_value

        return band_arr.astype(np.float32)

    def _open_mask_lt_4_0(
        self, mask_id: Union[str, S2GmlMasks], band: Union[BandNames, str] = None
    ) -> gpd.GeoDataFrame:
        """
        Open S2 mask (GML files stored in QI_DATA/qi) as :code:`gpd.GeoDataFrame`.

        Masks than can be called that way are:

        - :code:`TECQUA`: Technical quality mask
        - :code:`SATURA`: Saturated Pixels
        - :code:`NODATA`: Pixel nodata (inside the detectors)
        - :code:`DETFOO`: Detectors footprint -> used to process nodata outside the detectors
        - :code:`DEFECT`: Defective pixels
        - :code:`CLOUDS`, **only with :code:`00` as a band !**

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> from eoreader.bands import *
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader.open(r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip")
            >>> prod.open_mask("NODATA", GREEN)
            Empty GeoDataFrame
            Columns: [geometry]
            Index: []
            >>> prod.open_mask("SATURA", GREEN)
            Empty GeoDataFrame
            Columns: [geometry]
            Index: []
            >>> prod.open_mask("DETFOO", GREEN)
                                    gml_id  ...                                           geometry
            0  detector_footprint-B03-02-0  ...  POLYGON Z ((199980.000 4500000.000 0.000, 1999...
            1  detector_footprint-B03-03-1  ...  POLYGON Z ((222570.000 4500000.000 0.000, 2225...
            2  detector_footprint-B03-05-2  ...  POLYGON Z ((273050.000 4500000.000 0.000, 2730...
            3  detector_footprint-B03-07-3  ...  POLYGON Z ((309770.000 4453710.000 0.000, 3097...
            4  detector_footprint-B03-04-4  ...  POLYGON Z ((248080.000 4500000.000 0.000, 2480...
            5  detector_footprint-B03-06-5  ...  POLYGON Z ((297980.000 4500000.000 0.000, 2979...
            [6 rows x 3 columns]

        Args:
            mask_id (Union[str, S2GmlMasks]): Mask name, such as DEFECT, NODATA, SATURA...
            band (Union[BandNames, str]): Band number as an SpectralBandNames or str (for clouds: 00)

        Returns:
            gpd.GeoDataFrame: Mask as a vector
        """
        # Check inputs
        mask_id = S2GmlMasks.from_value(mask_id)
        if mask_id == S2GmlMasks.CLOUDS:
            band = "00"

        # Get QI_DATA path
        if isinstance(band, BandNames):
            band_name = self.bands[band].id
        else:
            band_name = band

        tmp_dir = tempfile.TemporaryDirectory()
        try:
            if self.is_archived:
                # Open the zip file
                # WE DON'T KNOW WHY BUT DO NOT USE path.read_archived_vector HERE !!!
                with zipfile.ZipFile(self.path, "r") as zip_ds:
                    filenames = [f.filename for f in zip_ds.filelist]
                    regex = re.compile(
                        f"{self._get_qi_folder()}.*{mask_id.value}_B{band_name}.gml"
                    )
                    mask_path = zip_ds.extract(
                        list(filter(regex.match, filenames))[0], tmp_dir.name
                    )
            else:
                # Get mask path
                mask_path = path.get_file_in_dir(
                    self.path,
                    f"{self._get_qi_folder()}/*{mask_id.value}_B{band_name}.gml",
                    exact_name=True,
                )

            # Read vector
            mask = vectors.read(mask_path, crs=self.crs())

        except Exception as ex:
            raise InvalidProductError(ex) from ex

        finally:
            tmp_dir.cleanup()

        return mask

    def _open_mask_gt_4_0(
        self,
        mask_id: Union[str, S2Jp2Masks],
        band: Union[BandNames, str] = None,
        pixel_size: float = None,
        size: Union[list, tuple] = None,
        **kwargs,
    ) -> xr.DataArray:
        """
        Open S2 mask (jp2 files stored in QI_DATA) as raster.

        Masks than can be called that way are:

        - :code:`DETFOO`: Detectors footprint -> used to process nodata outside the detectors
        - :code:`QUALIT`: TECQUA, DEFECT, NODATA, SATURA, CLOLOW merged
        - :code:`CLASSI`: CLOUDS and SNOICE **only with :code:`00` as a band !**

        Args:
            mask_id (Union[str, S2GmlMasks]): Mask ID
            band (Union[BandNames, str]): Band number as an SpectralBandNames or str (for clouds: 00)
            pixel_size (int): Band pixel size in meters
            size (Union[tuple, list]): Size of the array (width, height). Not used if pixel_size is provided.

        Returns:
            gpd.GeoDataFrame: Mask as a DataArray
        """
        # Check inputs
        mask_id = S2Jp2Masks.from_value(mask_id)
        if mask_id == S2Jp2Masks.CLOUDS:
            band = "00"

        # Get QI_DATA path
        if isinstance(band, BandNames):
            band_id = self.bands[band].id
        else:
            band_id = band

        if self.is_archived:
            mask_path = path.get_archived_rio_path(
                self.path, f"{self._get_qi_folder()}.*{mask_id.value}_B{band_id}.jp2"
            )
        else:
            # Get mask path
            mask_path = path.get_file_in_dir(
                self.path,
                f"{self._get_qi_folder()}/*{mask_id.value}_B{band_id}.jp2",
                exact_name=True,
            )

        # Read mask
        mask = utils.read(
            mask_path,
            pixel_size=pixel_size,
            size=size,
            resampling=Resampling.nearest,
            as_type=np.uint8,
            masked=False,
            **kwargs,
        )

        return mask

    def _manage_invalid_pixels(
        self, band_arr: xr.DataArray, band: BandNames, **kwargs
    ) -> xr.DataArray:
        """
        Manage invalid pixels (Nodata, saturated, defective...)
        See there: https://sentinel.esa.int/documents/247904/349490/S2_MSI_Product_Specification.pdf

        Args:
            band_arr (xr.DataArray): Band array
            band (BandNames): Band name as an SpectralBandNames
            kwargs: Other arguments used to load bands

        Returns:
            xr.DataArray: Cleaned band array
        """
        if self._processing_baseline < 4.0:
            return self._manage_invalid_pixels_lt_4_0(band_arr, band, **kwargs)
        else:
            # return band_arr
            return self._manage_invalid_pixels_gt_4_0(band_arr, band, **kwargs)

    def _manage_nodata(
        self, band_arr: xr.DataArray, band: BandNames, **kwargs
    ) -> xr.DataArray:
        """
        Manage only nodata pixels

        Args:
            band_arr (xr.DataArray): Band array
            band (BandNames): Band name as an SpectralBandNames
            kwargs: Other arguments used to load bands

        Returns:
            xr.DataArray: Cleaned band array
        """
        if self._processing_baseline < 4.0:
            return self._manage_nodata_lt_4_0(band_arr, band, **kwargs)
        else:
            return self._manage_nodata_gt_4_0(band_arr, band, **kwargs)

    def _manage_invalid_pixels_lt_4_0(
        self, band_arr: xr.DataArray, band: BandNames, **kwargs
    ) -> xr.DataArray:
        """
        Manage invalid pixels (Nodata, saturated, defective...)
        See there: https://sentinel.esa.int/documents/247904/349490/S2_MSI_Product_Specification.pdf

        Args:
            band_arr (xr.DataArray): Band array
            band (BandNames): Band name as an SpectralBandNames
            kwargs: Other arguments used to load bands

        Returns:
            xr.DataArray: Cleaned band array
        """
        # Get detector footprint to deduce the outside nodata
        nodata_det = self._open_mask_lt_4_0(
            S2GmlMasks.FOOTPRINT, band
        )  # Detector nodata, -> pixels that are outside the detectors

        if len(nodata_det) > 0:
            # Rasterize nodata
            mask = features.rasterize(
                nodata_det.geometry,
                out_shape=(band_arr.rio.height, band_arr.rio.width),
                fill=self._mask_true,  # Outside detector = nodata (inverted compared to the usual)
                default_value=self._mask_false,  # Inside detector = not nodata
                transform=transform.from_bounds(
                    *band_arr.rio.bounds(), band_arr.rio.width, band_arr.rio.height
                ),
                dtype=np.uint8,
            )
        else:
            # Manage empty geometry: nodata is 0
            LOGGER.warning(
                "Empty detector footprint (DETFOO) vector. Nodata will be set where the pixels are null."
            )
            s2_nodata = 0
            mask = np.where(band_arr == s2_nodata, 1, 0).astype(np.uint8)

        #  Load masks and merge them into the nodata
        nodata_pix = self._open_mask_lt_4_0(
            S2GmlMasks.NODATA, band
        )  # Pixel nodata, not pixels that are outside the detectors !!!
        if len(nodata_pix) > 0:
            # Discard pixels corrected during crosstalk
            nodata_pix = nodata_pix[nodata_pix.gml_id == "QT_NODATA_PIXELS"]
        nodata_pix = pd.concat(
            [nodata_pix, self._open_mask_lt_4_0(S2GmlMasks.DEFECT, band)]
        )
        nodata_pix = pd.concat(
            [nodata_pix, self._open_mask_lt_4_0(S2GmlMasks.SATURATION, band)]
        )

        # Technical quality mask
        tecqua = self._open_mask_lt_4_0(S2GmlMasks.QUALITY, band)
        if len(tecqua) > 0:
            # Do not take into account ancillary data
            tecqua = tecqua[tecqua.gml_id.isin(["MSI_LOST", "MSI_DEG"])]
        nodata_pix = pd.concat([nodata_pix, tecqua])

        if len(nodata_pix) > 0:
            # Rasterize mask
            mask_pix = features.rasterize(
                nodata_pix.geometry,
                out_shape=(band_arr.rio.height, band_arr.rio.width),
                fill=self._mask_false,  # Outside vector
                default_value=self._mask_true,  # Inside vector
                transform=transform.from_bounds(
                    *band_arr.rio.bounds(), band_arr.rio.width, band_arr.rio.height
                ),
                dtype=np.uint8,
            )

            mask[mask_pix] = self._mask_true

        return self._set_nodata_mask(band_arr, mask)

    def _manage_invalid_pixels_gt_4_0(
        self, band_arr: xr.DataArray, band: BandNames, **kwargs
    ) -> xr.DataArray:
        """
        Manage invalid pixels (Nodata, saturated, defective...)
        See there:
        https://sentinels.copernicus.eu/documents/247904/685211/Sentinel-2-Products-Specification-Document-14_8.pdf

        Args:
            band_arr (xr.DataArray): Band array
            band (BandNames): Band name as an SpectralBandNames
            kwargs: Other arguments used to load bands

        Returns:
            xr.DataArray: Cleaned band array
        """
        # Get detector footprint to deduce the outside nodata
        nodata = self._open_mask_gt_4_0(
            S2Jp2Masks.FOOTPRINT,
            band,
            size=(band_arr.rio.width, band_arr.rio.height),
            **kwargs,
        ).data

        nodata = np.where(nodata == 0, 1, 0).astype(np.uint8)

        # Manage quality mask
        # TODO: Optimize it -> very slow (why ?)
        # Technical quality mask: Only keep MSI_LOST (band 3) and MSI_DEG (band 4)
        # Defective pixels (band 5)
        # Nodata pixels (band 6)
        # Saturated pixels (band 8)
        quality = self._open_mask_gt_4_0(
            S2Jp2Masks.QUALITY,
            band,
            size=(band_arr.rio.width, band_arr.rio.height),
            indexes=[3, 4, 5, 6, 8],
            **kwargs,
        ).data

        # Compute mask
        mask = (nodata + np.sum(quality, axis=0)) > 0

        return self._set_nodata_mask(band_arr, mask)

    def _manage_nodata_lt_4_0(
        self, band_arr: xr.DataArray, band: BandNames, **kwargs
    ) -> xr.DataArray:
        """
        Manage only nodata
        See there: https://sentinel.esa.int/documents/247904/349490/S2_MSI_Product_Specification.pdf

        Args:
            band_arr (xr.DataArray): Band array
            band (BandNames): Band name as an SpectralBandNames
            kwargs: Other arguments used to load bands

        Returns:
            xr.DataArray: Cleaned band array
        """
        # Get detector footprint to deduce the outside nodata
        nodata_det = self._open_mask_lt_4_0(
            S2GmlMasks.FOOTPRINT, band
        )  # Detector nodata, -> pixels that are outside the detectors

        if len(nodata_det) > 0:
            # Rasterize nodata
            mask = features.rasterize(
                nodata_det.geometry,
                out_shape=(band_arr.rio.height, band_arr.rio.width),
                fill=self._mask_true,  # Outside detector = nodata (inverted compared to the usual)
                default_value=self._mask_false,  # Inside detector = not nodata
                transform=transform.from_bounds(
                    *band_arr.rio.bounds(), band_arr.rio.width, band_arr.rio.height
                ),
                dtype=np.uint8,
            )
        else:
            # Manage empty geometry: nodata is 0
            LOGGER.warning(
                "Empty detector footprint (DETFOO) vector. Nodata will be set where the pixels are null."
            )
            s2_nodata = 0
            mask = np.where(band_arr == s2_nodata, 1, 0).astype(np.uint8)

        return self._set_nodata_mask(band_arr, mask)

    def _manage_nodata_gt_4_0(
        self, band_arr: xr.DataArray, band: BandNames, **kwargs
    ) -> xr.DataArray:
        """
        Manage only nodata
        See there: https://sentinel.esa.int/documents/247904/349490/S2_MSI_Product_Specification.pdf

        Args:
            band_arr (xr.DataArray): Band array
            band (BandNames): Band name as an SpectralBandNames
            kwargs: Other arguments used to load bands

        Returns:
            xr.DataArray: Cleaned band array
        """
        # Get detector footprint to deduce the outside nodata
        nodata = self._open_mask_gt_4_0(
            S2Jp2Masks.FOOTPRINT,
            band,
            size=(band_arr.rio.width, band_arr.rio.height),
            **kwargs,
        ).data

        nodata = np.where(nodata == 0, 1, 0).astype(np.uint8)

        return self._set_nodata_mask(band_arr, nodata)

    def _load_bands(
        self,
        bands: list,
        pixel_size: float = None,
        size: Union[list, tuple] = None,
        **kwargs,
    ) -> dict:
        """
        Load bands as numpy arrays with the same pixel size (and same metadata).

        Args:
            bands (list): List of the wanted bands
            pixel_size (float): Band pixel size in meters
            size (Union[tuple, list]): Size of the array (width, height). Not used if pixel_size is provided.
            kwargs: Other arguments used to load bands
        Returns:
            dict: Dictionary {band_name, band_xarray}
        """
        # Return empty if no band are specified
        if not bands:
            return {}

        if pixel_size is None and size is not None:
            pixel_size = self._pixel_size_from_img_size(size)
        band_paths = self.get_band_paths(bands, pixel_size=pixel_size, **kwargs)

        # Open bands and get array (resampled if needed)
        band_arrays = self._open_bands(
            band_paths, pixel_size=pixel_size, size=size, **kwargs
        )

        return band_arrays

    def _get_condensed_name(self) -> str:
        """
        Get S2 products condensed name ({date}_S2_{tile}_{product_type}_{generation_time}).

        Returns:
            str: Condensed name
        """
        # Used to make the difference between 2 products acquired on the same tile at the same date but cut differently
        # Sentinel-2 generation time: "%Y%m%dT%H%M%S" -> save only %H%M%S
        gen_time = self.split_name[-1].split("T")[-1]

        # Force S2 as constellation name for S2_SIN to work
        return f"{self.get_datetime()}_S2_{self.tile_name}_{self.product_type.name}_{gen_time}"

    @cache
    def get_mean_sun_angles(self) -> (float, float):
        """
        Get Mean Sun angles (Azimuth and Zenith angles)

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.get_mean_sun_angles()
            (149.148155074489, 32.6627897525474)

        Returns:
            (float, float): Mean Azimuth and Zenith angle
        """
        try:
            # Read metadata
            root, _ = self.read_mtd()

            try:
                mean_sun_angles = root.find(".//Mean_Sun_Angle")
                zenith_angle = float(mean_sun_angles.findtext("ZENITH_ANGLE"))
                azimuth_angle = float(mean_sun_angles.findtext("AZIMUTH_ANGLE"))
            except TypeError:
                raise InvalidProductError(
                    "Azimuth or Zenith angles not found in metadata!"
                )
        except InvalidProductError as ex:
            LOGGER.warning(f"{ex}: setting sun angles to (0, 0).")
            azimuth_angle = 0.0
            zenith_angle = 0.0

        return azimuth_angle, zenith_angle

    @cache
    def _read_mtd(self) -> (etree._Element, dict):
        """
        Read metadata and outputs the metadata XML root and its namespaces as a dict

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.read_mtd()
            (<Element {https://psd-14.sentinel2.eo.esa.int/PSD/S2_PDI_Level-2A_Tile_Metadata.xsd}Level-2A_Tile_ID at ...>,
            {'nl': '{https://psd-14.sentinel2.eo.esa.int/PSD/S2_PDI_Level-2A_Tile_Metadata.xsd}'})

        Returns:
            (etree._Element, dict): Metadata XML root and its namespaces
        """
        if self._is_sinergise:
            mtd_from_path = "metadata.xml"
            mtd_archived = r"metadata\.xml"
        else:
            mtd_from_path = "GRANULE/*/MTD*.xml"
            mtd_archived = r"GRANULE.*MTD.*\.xml"

        return self._read_mtd_xml(mtd_from_path, mtd_archived)

    @cache
    def read_datatake_mtd(self) -> (etree._Element, dict):
        """
        Read datatake metadata and outputs the metadata XML root and its namespaces as a dict
        (datatake metadata is the file in the root directory named :code:`MTD_MSI(L1C/L2A).xml`)

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.read_mtd()
            (<Element {https://psd-14.sentinel2.eo.esa.int/PSD/S2_PDI_Level-2A_Tile_Metadata.xsd}Level-2A_Tile_ID at ...>,
            {'nl': '{https://psd-14.sentinel2.eo.esa.int/PSD/S2_PDI_Level-2A_Tile_Metadata.xsd}'})

        Returns:
            (etree._Element, dict): Metadata XML root and its namespaces
        """
        mtd_from_path = "MTD_MSI*.xml"
        mtd_archived = r"MTD_MSI.*\.xml"

        return self._read_mtd_xml(mtd_from_path, mtd_archived)

    def _has_cloud_band(self, band: BandNames) -> bool:
        """
        Does this product has the specified cloud band ?
        https://sentinels.copernicus.eu/web/sentinel/technical-guides/sentinel-2-msi/level-1c/cloud-masks
        """
        if band == SHADOWS:
            has_band = False
        else:
            has_band = True
        return has_band

    def _open_clouds_lt_4_0(
        self,
        bands: list,
        pixel_size: float = None,
        size: Union[list, tuple] = None,
        **kwargs,
    ) -> dict:
        """
        Load cloud files as xarrays.

        Read S2 cloud mask .GML files (both valid for L2A and L1C products).
        https://sentinels.copernicus.eu/web/sentinel/technical-guides/sentinel-2-msi/level-1c/cloud-masks

        Args:
            bands (list): List of the wanted bands
            pixel_size (int): Band pixel size in meters
            size (Union[tuple, list]): Size of the array (width, height). Not used if pixel_size is provided.
            kwargs: Additional arguments
        Returns:
            dict: Dictionary {band_name, band_xarray}
        """
        band_dict = {}

        if bands:
            cloud_vec = self._open_mask_lt_4_0(S2GmlMasks.CLOUDS)

            # Open a bands to mask it
            def_band = self._read_band(
                self.get_default_band_path(),
                self.get_default_band(),
                pixel_size=pixel_size,
                size=size,
            )
            nodata = np.where(np.isnan(def_band), 1, 0)

            for band in bands:
                if band == ALL_CLOUDS:
                    cloud = self._rasterize(def_band, cloud_vec, nodata)
                elif band == CIRRUS:
                    try:
                        cirrus = cloud_vec[cloud_vec.maskType == "CIRRUS"]
                    except AttributeError:
                        # No masktype -> empty
                        cirrus = gpd.GeoDataFrame(geometry=[], crs=cloud_vec.crs)
                    cloud = self._rasterize(def_band, cirrus, nodata)
                elif band == CLOUDS:
                    try:
                        clouds = cloud_vec[cloud_vec.maskType == "OPAQUE"]
                    except AttributeError:
                        # No masktype -> empty
                        clouds = gpd.GeoDataFrame(geometry=[], crs=cloud_vec.crs)
                    cloud = self._rasterize(def_band, clouds, nodata)
                elif band == RAW_CLOUDS:
                    cloud = self._rasterize(def_band, cloud_vec, nodata)
                else:
                    raise InvalidTypeError(
                        f"Non existing cloud band for Sentinel-2: {band}"
                    )

                # Rename
                band_name = to_str(band)[0]

                # Multi bands -> do not change long name
                if band != RAW_CLOUDS:
                    cloud.attrs["long_name"] = band_name
                band_dict[band] = cloud.rename(band_name).astype(np.float32)

        return band_dict

    def _open_clouds_gt_4_0(
        self,
        bands: list,
        pixel_size: float = None,
        size: Union[list, tuple] = None,
        **kwargs,
    ) -> dict:
        """
        Load cloud files as xarrays.

        Read S2 cloud mask .JP2 files (both valid for L2A and L1C products).
        https://sentinels.copernicus.eu/documents/247904/685211/Sentinel-2-Products-Specification-Document-14_8.pdf

        Args:
            bands (list): List of the wanted bands
            pixel_size (int): Band pixel size in meters
            size (Union[tuple, list]): Size of the array (width, height). Not used if pixel_size is provided.
            kwargs: Additional arguments
        Returns:
            dict: Dictionary {band_name, band_xarray}
        """
        band_dict = {}

        if bands:
            cloud_vec = self._open_mask_gt_4_0(
                S2Jp2Masks.CLOUDS,
                "00",
                pixel_size=pixel_size,
                size=size,
                **kwargs,
            )

            for band in bands:
                if band == ALL_CLOUDS:
                    cloud = cloud_vec[0, :, :] | cloud_vec[1, :, :]
                elif band == CIRRUS:
                    cloud = cloud_vec[1, :, :]  # CIRRUS = band 2
                elif band == CLOUDS:
                    cloud = cloud_vec[0, :, :]  # OPAQUE = band 1
                elif band == RAW_CLOUDS:
                    cloud = cloud_vec
                else:
                    raise InvalidTypeError(
                        f"Non existing cloud band for Sentinel-2: {band}"
                    )

                if len(cloud.shape) == 2:
                    cloud = cloud.expand_dims(dim="band", axis=0)

                # Rename
                band_name = to_str(band)[0]

                # Multi bands -> do not change long name
                if band != RAW_CLOUDS:
                    cloud.attrs["long_name"] = band_name

                band_dict[band] = cloud.rename(band_name).astype(np.float32)

        return band_dict

    def _open_clouds(
        self,
        bands: list,
        pixel_size: float = None,
        size: Union[list, tuple] = None,
        **kwargs,
    ) -> dict:
        """
        Load cloud files as xarrays.

        Read S2 cloud mask .GML files (both valid for L2A and L1C products).
        https://sentinels.copernicus.eu/web/sentinel/technical-guides/sentinel-2-msi/level-1c/cloud-masks

        Args:
            bands (list): List of the wanted bands
            pixel_size (int): Band pixel size in meters
            size (Union[tuple, list]): Size of the array (width, height). Not used if pixel_size is provided.
            kwargs: Additional arguments
        Returns:
            dict: Dictionary {band_name, band_xarray}
        """
        if self._processing_baseline < 4.0:
            return self._open_clouds_lt_4_0(bands, pixel_size, size, **kwargs)
        else:
            return self._open_clouds_gt_4_0(bands, pixel_size, size, **kwargs)

    def _rasterize(
        self, xds: xr.DataArray, geometry: gpd.GeoDataFrame, nodata: np.ndarray
    ) -> xr.DataArray:
        """
        Rasterize a vector on a memory dataset

        Args:
            xds (xr.DataArray): Array
            geometry (gpd.GeoDataFrame): Geometry to rasterize
            nodata (np.ndarray): Nodata mask

        Returns:
            xr.DataArray: Rasterized vector
        """
        if not geometry.empty:
            # Just in case
            if geometry.crs != xds.rio.crs:
                geometry = geometry.to_crs(xds.rio.crs)

            # Rasterize mask
            cond = features.rasterize(
                geometry.geometry,
                out_shape=(xds.rio.height, xds.rio.width),
                fill=self._mask_false,  # Pixels outside mask
                default_value=self._mask_true,  # Pixels inside mask
                transform=transform.from_bounds(
                    *xds.rio.bounds(), xds.rio.width, xds.rio.height
                ),
                dtype=np.uint8,
            )
            cond = np.expand_dims(cond, axis=0)

        else:
            # If empty geometry, just
            cond = np.full(
                shape=(xds.rio.count, xds.rio.height, xds.rio.width),
                fill_value=self._mask_false,
                dtype=np.uint8,
            )
        return self._create_mask(xds, cond, nodata)

    def _l2ap_geocode_data(self, l2ap_path: AnyPathType) -> (Affine, int, int, CRS):
        """
        Geocode L2Ap data.

        Args:
            l2ap_path (AnyPathType): Band path to be geocoded

        Returns:
            (Affine, int, int, CRS): Transform, width, height and CRS of the band
        """
        try:
            if isinstance(l2ap_path, str):
                l2ap_path = AnyPath(l2ap_path)

            # Read metadata
            root, ns = self.read_mtd()

            # Determine wanted resolution
            if "10m" in l2ap_path.name:
                res = 10
            elif "20m" in l2ap_path.name:
                res = 20
            else:
                res = 60

            # Open size
            width = int(root.findtext(f".//Size[@resolution='{res}']/NCOLS"))
            height = int(root.findtext(f".//Size[@resolution='{res}']/NROWS"))

            # Open upper-left corner
            ulx = float(root.findtext(f".//Geoposition[@resolution='{res}']/ULX"))
            uly = float(root.findtext(f".//Geoposition[@resolution='{res}']/ULY"))

            # Create transform
            tf = transform.from_origin(ulx, uly, res, res)
        except InvalidProductError as ex:
            raise InvalidProductError(f"{ex}: cannot geocode the bands!")

        return tf, width, height, self.crs()

    @cache
    def default_transform(self, **kwargs) -> (Affine, int, int, CRS):
        """
        Returns default transform data of the default band (UTM),
        as the :code:`rasterio.warp.calculate_default_transform` does:
        - transform
        - width
        - height
        - crs

        Args:
            kwargs: Additional arguments
        Returns:
            Affine, int, int, CRS: transform, width, height, CRS

        """
        if self._processing_baseline < 2.07:
            default_path = self.get_default_band_path(**kwargs)
            return self._l2ap_geocode_data(default_path)
        else:
            return super().default_transform()

    @cache
    def get_cloud_cover(self) -> float:
        """
        Get cloud cover as given in the metadata

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.get_cloud_cover()
            55.5

        Returns:
            float: Cloud cover as given in the metadata
        """
        # Get the cloud cover
        try:
            # Get MTD XML file
            root, nsmap = self.read_mtd()
            cc = float(root.findtext(".//CLOUDY_PIXEL_PERCENTAGE"))
        except (InvalidProductError, TypeError):
            LOGGER.warning(
                "CLOUDY_PIXEL_PERCENTAGE not found in metadata! Cloud coverage set to 0."
            )
            cc = 0

        return cc

    def get_quicklook_path(self) -> str:
        """
        Get quicklook path if existing (some providers are providing one quicklook, such as creodias)

        Returns:
            str: Quicklook path
        """
        quicklook_path = None
        try:
            if self.is_archived:
                quicklook_path = self.path / path.get_archived_path(
                    self.path, file_regex=r".*ql\.jpg"
                )
            else:
                quicklook_path = next(self.path.glob("**/*ql.jpg"))
        except (StopIteration, FileNotFoundError):
            try:
                if self.is_archived:
                    quicklook_path = self.path / path.get_archived_path(
                        self.path, file_regex=r".*preview\.jpg"
                    )
                else:
                    quicklook_path = next(self.path.glob("**/preview.jpg"))
            except (StopIteration, FileNotFoundError):
                # Use the PVI
                try:
                    if self.is_archived:
                        quicklook_path = path.get_archived_rio_path(
                            self.path, file_regex=r".*PVI\.jp2"
                        )
                    else:
                        quicklook_path = next(self.path.glob("**/*PVI.jp2"))
                except (StopIteration, FileNotFoundError):
                    LOGGER.warning(f"No quicklook found in {self.condensed_name}")

        if quicklook_path is not None:
            quicklook_path = str(quicklook_path)

        return quicklook_path

    @cache
    def get_orbit_direction(self) -> OrbitDirection:
        """
        Get cloud cover as given in the metadata

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.get_orbit_direction().value
            "DESCENDING"

        Returns:
            OrbitDirection: Orbit direction (ASCENDING/DESCENDING)
        """
        try:
            # Get MTD XML file
            root, _ = self.read_datatake_mtd()

            # Get the orbit direction
            try:
                od = OrbitDirection.from_value(
                    root.findtext(".//SENSING_ORBIT_DIRECTION")
                )

            except TypeError:
                raise InvalidProductError(
                    "SENSING_ORBIT_DIRECTION not found in metadata!"
                )
        except InvalidProductError:
            od = OrbitDirection.DESCENDING

        return od


class S2StacProduct(StacProduct, S2Product):
    def __init__(
        self,
        product_path: AnyPathStrType = None,
        archive_path: AnyPathStrType = None,
        output_path: AnyPathStrType = None,
        remove_tmp: bool = False,
        **kwargs,
    ) -> None:
        self.kwargs = kwargs
        """Custom kwargs"""

        # Copy the kwargs
        super_kwargs = kwargs.copy()

        # Get STAC Item
        self.item = None
        """ STAC Item of the product """
        self.item = super_kwargs.pop("item", None)
        if self.item is None:
            try:
                self.item = Item.from_file(product_path)
            except TypeError:
                raise InvalidProductError(
                    "You should either fill 'product_path' or 'item'."
                )

        self.default_clients = [
            planetary_computer,
            S3Client(region_name="us-west-2"),  # Element84, EarthData
            S3Client(requester_pays=True, region_name="eu-central-1"),  # Sinergise
            # Not yet handled
            # HttpClient(ClientSession(base_url="https://landsatlook.usgs.gov", auth=BasicAuth(login="", password="")))
        ]
        self.clients = super_kwargs.pop("client", self.default_clients)

        if product_path is None:
            # Canonical link is always the second one
            # TODO: check if ok
            product_path = AnyPath(self.item.links[1].target).parent

        # Initialization from the super class
        super().__init__(product_path, archive_path, output_path, remove_tmp, **kwargs)

    def _pre_init(self, **kwargs) -> None:
        """
        Function used to pre_init the products
        (setting needs_extraction and so on)
        """
        self._raw_units = RawUnits.REFL
        self._has_cloud_cover = True
        self._use_filename = False
        self.needs_extraction = False

        # Pre init done by the super class
        super(S2Product, self)._pre_init(**kwargs)

    def _get_path(self, file_id: str, ext="tif") -> str:
        """
        Get either the archived path of the normal path of a tif file

        Args:
            band_id (str): Band ID

        Returns:
            AnyPathType: band path
        """
        if file_id.lower() in self.item.assets:
            asset_name = file_id.lower()
        elif file_id in [band.id for band in self.bands.values() if band is not None]:
            band_name = [
                band_name
                for band_name, band in self.bands.items()
                if band is not None and f"{band.id}" == file_id
            ][0]
            asset_name = EOREADER_STAC_MAP[band_name].value
        else:
            try:
                asset_name = difflib.get_close_matches(
                    file_id, self.item.assets.keys(), cutoff=0.5, n=1
                )[0]
            except Exception:
                raise FileNotFoundError(
                    f"Impossible to find an asset in {list(self.item.assets.keys())} close enough to '{file_id}'"
                )

        return self.sign_url(self.item.assets[asset_name].href)

    def get_band_paths(
        self, band_list: list, pixel_size: float = None, **kwargs
    ) -> dict:
        """
        Return the paths of required bands.

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> from eoreader.bands import *
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.get_band_paths([GREEN, RED])
            {
                <SpectralBandNames.GREEN: 'GREEN'>: 'zip+file://S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip!/S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE/GRANULE/L1C_T30TTK_A027018_20200824T111345/IMG_DATA/T30TTK_20200824T110631_B03.jp2',
                <SpectralBandNames.RED: 'RED'>: 'zip+file://S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip!/S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE/GRANULE/L1C_T30TTK_A027018_20200824T111345/IMG_DATA/T30TTK_20200824T110631_B04.jp2'
            }

        Args:
            band_list (list): List of the wanted bands
            pixel_size (float): Band pixel size
            kwargs: Other arguments used to load bands

        Returns:
            dict: Dictionary containing the path of each queried band
        """
        band_paths = {}
        for band in band_list:
            # Get clean band path
            clean_band = self._get_clean_band_path(
                band, pixel_size=pixel_size, **kwargs
            )
            if clean_band.is_file():
                band_paths[band] = clean_band
            else:
                band_id = self.bands[band].id
                try:
                    band_paths[band] = self._get_path(band_id)
                except (FileNotFoundError, IndexError) as ex:
                    raise InvalidProductError(
                        f"Non existing {band} ({band_id}) band for {self.path}"
                    ) from ex

        return band_paths

    @cache
    def read_datatake_mtd(self) -> (etree._Element, dict):
        """
        Read datatake metadata and outputs the metadata XML root and its namespaces as a dict
        (datatake metadata is the file in the root directory named :code:`MTD_MSI(L1C/L2A).xml`)

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.read_mtd()
            (<Element {https://psd-14.sentinel2.eo.esa.int/PSD/S2_PDI_Level-2A_Tile_Metadata.xsd}Level-2A_Tile_ID at ...>,
            {'nl': '{https://psd-14.sentinel2.eo.esa.int/PSD/S2_PDI_Level-2A_Tile_Metadata.xsd}'})

        Returns:
            (etree._Element, dict): Metadata XML root and its namespaces
        """
        return self._read_mtd_xml_stac(self._get_path("product-metadata"))

    def _read_mtd(self) -> (etree._Element, dict):
        """
        Read Landsat metadata as:

         - :code:`pandas.DataFrame` whatever its collection is (by default for collection 1)
         - XML root + its namespace if the product is retrieved from the 2nd collection (by default for collection 2)

        Args:
            force_pd (bool): If collection 2, return a pandas.DataFrame instead of an XML root + namespace
        Returns:
            Tuple[Union[pd.DataFrame, etree._Element], dict]:
                Metadata as a Pandas.DataFrame or as (etree._Element, dict): Metadata XML root and its namespaces
        """
        return self._read_mtd_xml_stac(self._get_path("granule-metadata"))

    def get_quicklook_path(self) -> str:
        """
        Get quicklook path if existing.

        Returns:
            str: Quicklook path
        """
        return self._get_path("preview")



# -*- coding: utf-8 -*-
# Copyright 2024, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of eoreader project
#     https://github.com/sertit/eoreader
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" STAC KEYWORDS """
from sertit.misc import ListEnum

__all__ = [
    "ID",
    "GSD",
    "TITLE",
    "DATETIME",
    "GEOMETRY",
    "BBOX",
    "STAC_EXTENSIONS",
    "NA",
    "CONSTELLATION",
    "PLATFORM",
    "INSTRUMENTS",
    "EO_BANDS",
    "EO_CC",
    "NAME",
    "COMMON_NAME",
    "DESCRIPTION",
    "CENTER_WV",
    "WV_MIN",
    "WV_MAX",
    "FWHM",
    "SOLAR_ILLUMINATION",
    "StacCommonNames",
    "ASSET_ROLE",
    "REFLECTANCE",
    "BT",
    "SATURATION",
    "CLOUD",
    "CLOUD_SHADOW",
    "AMPLITUDE",
    "INTENSITY",
    "COHERENCE",
    "PROJ_BBOX",
    "PROJ_EPSG",
    "PROJ_WKT",
    "PROJ_SHAPE",
    "PROJ_GEOMETRY",
    "PROJ_TRANSFORM",
    "PROJ_CENTROID",
    "VIEW_AZIMUTH",
    "VIEW_OFF_NADIR",
    "VIEW_SUN_AZIMUTH",
    "VIEW_INCIDENCE_ANGLE",
    "VIEW_SUN_ELEVATION",
]

# MISC

ID = "id"
GSD = "gsd"
TITLE = "title"
GEOMETRY = "geometry"
BBOX = "bbox"
DATETIME = "datetime"
NA = "N/A"
CONSTELLATION = "constellation"
PLATFORM = "platform"
INSTRUMENTS = "instruments"
STAC_EXTENSIONS = "stac_extensions"

# ---------------- Electro-Optical Extension Specification ----------------

# Additional Field Information: https://github.com/stac-extensions/eo/#additional-field-information
EO_BANDS = "eo:bands"
EO_CC = "eo:cloud_cover"

# Band object: https://github.com/stac-extensions/eo/#band-object
NAME = "name"
COMMON_NAME = "common_name"
DESCRIPTION = "description"
CENTER_WV = "center_wavelength"
WV_MIN = "min_wavelength"
WV_MAX = "max_wavelength"
FWHM = "full_width_half_max"
SOLAR_ILLUMINATION = "solar_illumination"


# Band common names: https://github.com/stac-extensions/eo/#common-band-names
class StacCommonNames(ListEnum):
    COASTAL = "coastal"  # Band 1, Sentinel-2
    BLUE = "blue"  # Band 2, Sentinel-2
    GREEN = "green"  # Band 3, Sentinel-2
    RED = "red"  # Band 4, Sentinel-2
    YELLOW = "yellow"  # Band Oa07, Sentinel-3 OLCI
    PAN = "pan"  # Band 8, Landsat OLI
    RE = "rededge"  # Band 5,6,7, Sentinel-2
    NIR = "nir"  # Band 8, Sentinel-2
    NIR08 = "nir08"  # Band 8a, Sentinel-2
    NIR09 = "nir09"  # Band 9, Sentinel-2
    CIRRUS = "cirrus"  # Band 10, Sentinel-2
    SWIR16 = "swir16"  # Band 11, Sentinel-2
    SWIR22 = "swir22"  # Band 12, Sentinel-2
    LWIR = "lwir"  # Band 6, Landsat TM
    LWIR11 = "lwir11"  # Band 10, Landsat TIRS
    LWIR12 = "lwir12"  # Band 11, Landsat TIRS


# Asset roles: https://github.com/stac-extensions/eo/#best-practices
ASSET_ROLE = "asset_role"

# Optical
REFLECTANCE = "reflectance"
BT = "brightness_temperature"
SATURATION = "saturation"
CLOUD = "cloud"
CLOUD_SHADOW = "cloud-shadow"

# SAR
INTENSITY = "intensity"
AMPLITUDE = "amplitude"
COHERENCE = "coherence"

# ---------------- Projection Extension Specification ----------------

# https://github.com/stac-extensions/projection/
PROJ_EPSG = "proj:epsg"
PROJ_WKT = "proj:wkt2"
PROJ_GEOMETRY = "proj:geometry"
PROJ_BBOX = "proj:bbox"
PROJ_CENTROID = "proj:centroid"
PROJ_SHAPE = "proj:shape"
PROJ_TRANSFORM = "proj:transform"

# ---------------- View Extension Specification ----------------

# https://github.com/stac-extensions/view/
VIEW_OFF_NADIR = "view:off_nadir"
VIEW_INCIDENCE_ANGLE = "view:incidence_angle"
VIEW_AZIMUTH = "view:azimuth"
VIEW_SUN_AZIMUTH = "view:sun_azimuth"
VIEW_SUN_ELEVATION = "view:sun_elevation"



# -*- coding: utf-8 -*-
# Copyright 2024, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of eoreader project
#     https://github.com/sertit/eoreader
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
VHR (very high resoplution) super class.
See `here <https://earth.esa.int/eogateway/documents/20142/37627/DigitalGlobe-Standard-Imagery.pdf>`_
for more information.
"""
import logging
import os
from abc import abstractmethod
from typing import Union

import affine
import numpy as np
import rasterio
import xarray as xr
from rasterio import rpc
from rasterio import shutil as rio_shutil
from rasterio import warp
from rasterio.crs import CRS
from rasterio.enums import Resampling
from rasterio.vrt import WarpedVRT
from sertit import AnyPath, path, rasters, rasters_rio
from sertit.snap import MAX_CORES
from sertit.types import AnyPathStrType, AnyPathType

from eoreader import EOREADER_NAME, utils
from eoreader.bands import BandNames
from eoreader.env_vars import DEM_PATH, TILE_SIZE
from eoreader.exceptions import InvalidProductError
from eoreader.keywords import DEM_KW
from eoreader.products import OpticalProduct
from eoreader.utils import DEFAULT_TILE_SIZE

LOGGER = logging.getLogger(EOREADER_NAME)


class VhrProduct(OpticalProduct):
    """
    Super Class of VHR products.

    Implementing mechanisms for orthorectification, default transform...
    """

    def __init__(
        self,
        product_path: AnyPathStrType,
        archive_path: AnyPathStrType = None,
        output_path: AnyPathStrType = None,
        remove_tmp: bool = False,
        **kwargs,
    ) -> None:
        self.ortho_path = None
        """
        Orthorectified path.
        Can be set to use manually orthorectified or pansharpened data, especially useful for VHR data on steep terrain.
        """

        self._proj_prod_type = []

        self.band_combi = None
        """
        Band combination, i.e. PAN, PMS, MS...
        """

        # Order id (product name more or less)
        self._order_id = None

        # Resolutions
        self._pan_res = None
        self._ms_res = None

        self._job_id = None
        self._raw_nodata = 0

        # Initialization from the super class
        super().__init__(product_path, archive_path, output_path, remove_tmp, **kwargs)

    def _post_init(self, **kwargs) -> None:
        """
        Function used to post_init the products
        (setting sensor type, band names and so on)
        """
        # Job ID
        self._job_id = self._get_job_id()

        # Pre init done by the super class
        super()._post_init(**kwargs)

    def get_default_band_path(self, **kwargs) -> AnyPathType:
        """
        Get default band (:code:`GREEN` for optical data) path.

        .. WARNING:
            If you are using a non orthorectified product, this function will orthorectify the stack.
            To do so, you **MUST** provide a DEM through the EOREADER_DEM_PATH environment variable

        .. WARNING:
            If you are using a non-projected product, this function will reproject the stack.

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"IMG_PHR1B_PMS_001"
            >>> prod = Reader().open(path)
            >>> prod.get_default_band_path()
            'IMG_PHR1A_PMS_001/DIM_PHR1A_PMS_202005110231585_ORT_5547047101.XML'

        Args:
            kwargs: Additional arguments
        Returns:
            AnyPathType: Default band path
        """
        return self._get_default_utm_band(self.pixel_size, **kwargs)

    @abstractmethod
    def _get_raw_crs(self) -> CRS:
        """
        Get raw CRS of the tile

        Returns:
            rasterio.crs.CRS: CRS object
        """
        raise NotImplementedError

    def _get_ortho_path(self, **kwargs) -> AnyPathType:
        """
        Get the orthorectified path of the bands.

        Returns:
            AnyPathType: Orthorectified path
        """
        if self.product_type in self._proj_prod_type:
            ortho_name = f"{self.condensed_name}_ortho.tif"

            ortho_path, ortho_exists = self._get_out_path(ortho_name)
            if not ortho_exists:
                LOGGER.info(
                    "Manually orthorectified stack not given by the user. "
                    "Reprojecting whole stack, this may take a while. "
                    "(May be inaccurate on steep terrain, depending on the DEM pixel size)"
                )

                # Reproject and write on disk data
                dem_path = self._get_dem_path(**kwargs)
                with rasterio.open(str(self._get_tile_path())) as src:
                    if "rpcs" in kwargs:
                        rpcs = kwargs.pop("rpcs")
                    else:
                        rpcs = src.rpcs

                    if not rpcs:
                        raise InvalidProductError(
                            "Your projected VHR data doesn't have any RPC. "
                            "EOReader cannot orthorectify it!"
                        )

                    out_arr, meta = self._reproject(
                        src.read(), src.meta, rpcs, dem_path, **kwargs
                    )
                    rasters_rio.write(out_arr, meta, ortho_path, tags=src.tags())

        else:
            ortho_path = self._get_tile_path()

        return ortho_path

    def get_band_paths(
        self, band_list: list, pixel_size: float = None, **kwargs
    ) -> dict:
        """
        Return the paths of required bands.

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> from eoreader.bands import *
            >>> path = r"IMG_PHR1B_PMS_001"
            >>> prod = Reader().open(path)
            >>> prod.get_band_paths([GREEN, RED])
            {
                <SpectralBandNames.GREEN: 'GREEN'>:
                'IMG_PHR1A_PMS_001/DIM_PHR1A_PMS_202005110231585_ORT_5547047101.XML',
                <SpectralBandNames.RED: 'RED'>:
                'IMG_PHR1A_PMS_001/DIM_PHR1A_PMS_202005110231585_ORT_5547047101.XML'
            }

        Args:
            band_list (list): List of the wanted bands
            pixel_size (float): Band pixel size
            kwargs: Other arguments used to load bands

        Returns:
            dict: Dictionary containing the path of each queried band
        """
        if not self.ortho_path:
            self.ortho_path = self._get_ortho_path(**kwargs)

        # Processed path names
        band_paths = {}
        for band in band_list:
            # Get clean band path
            clean_band = self._get_clean_band_path(
                band, pixel_size=pixel_size, **kwargs
            )
            if clean_band.is_file():
                band_paths[band] = clean_band
            else:
                # First look for reprojected bands
                reproj_path = self._get_utm_band_path(
                    band=band.name, pixel_size=pixel_size
                )
                if not reproj_path.is_file():
                    # Then for original data
                    band_path = self.ortho_path
                else:
                    band_path = reproj_path

                band_paths[band] = band_path

        return band_paths

    def _get_dem_path(self, **kwargs) -> str:
        """
        Get DEM path

        Returns:
            str: DEM path

        """
        # Get DEM path
        dem_path = os.environ.get(DEM_PATH, kwargs.get(DEM_KW))
        if not dem_path:
            raise ValueError(
                f"As you are using a non orthorectified VHR product ({self.path}), "
                f"you must provide a valid DEM through the {DEM_PATH} environment variable"
            )
        else:
            if path.is_cloud_path(dem_path):
                raise ValueError(
                    "gdalwarp cannot process DEM stored on cloud with 'RPC_DEM' argument, "
                    "hence cloud-stored DEM cannot be used with non orthorectified DIMAP data."
                    f"(DEM: {dem_path}, DIMAP data: {self.name})"
                )

        return dem_path

    def _reproject(
        self, src_arr: np.ndarray, src_meta: dict, rpcs: rpc.RPC, dem_path, **kwargs
    ) -> (np.ndarray, dict):
        """
        Reproject using RPCs (cannot use another pixel size than src to ensure RPCs are valid)

        Args:
            src_arr (np.ndarray): Array to reproject
            src_meta (dict): Metadata
            rpcs (rpc.RPC): RPCs
            dem_path (str): DEM path

        Returns:
            (np.ndarray, dict): Reprojected array and its metadata
        """

        # Set RPC keywords
        LOGGER.debug(f"Orthorectifying data with {dem_path}")
        kwargs = {
            "RPC_DEM": dem_path,
            "RPC_DEM_MISSING_VALUE": 0,
            "OSR_USE_ETMERC": "YES",
            "BIGTIFF": "IF_NEEDED",
        }

        # Reproject
        # WARNING: may not give correct output pixel size
        out_arr, dst_transform = warp.reproject(
            src_arr,
            rpcs=rpcs,
            src_crs=self._get_raw_crs(),
            src_nodata=self._raw_nodata,
            dst_crs=self.crs(),
            dst_resolution=self.pixel_size,
            dst_nodata=self._raw_nodata,  # input data should be in integer
            num_threads=MAX_CORES,
            resampling=Resampling.bilinear,
            **kwargs,
        )
        # Get dims
        count, height, width = out_arr.shape

        # Update metadata
        meta = src_meta.copy()
        meta["transform"] = dst_transform
        meta["driver"] = "GTiff"
        meta["compress"] = "lzw"
        meta["nodata"] = self._raw_nodata
        meta["crs"] = self.crs()
        meta["width"] = width
        meta["height"] = height
        meta["count"] = count

        return out_arr, meta

    def _read_band(
        self,
        band_path: AnyPathType,
        band: BandNames = None,
        pixel_size: Union[tuple, list, float] = None,
        size: Union[list, tuple] = None,
        **kwargs,
    ) -> xr.DataArray:
        """
        Read band from disk.

        .. WARNING::
            Invalid pixels are not managed here

        Args:
            band_path (AnyPathType): Band path
            band (BandNames): Band to read
            pixel_size (Union[tuple, list, float]): Size of the pixels of the wanted band, in dataset unit (X, Y)
            size (Union[tuple, list]): Size of the array (width, height). Not used if pixel_size is provided.
            kwargs: Other arguments used to load bands
        Returns:
            xr.DataArray: Band xarray
        """
        with rasterio.open(str(band_path)) as dst:
            dst_crs = dst.crs

            # Compute pixel_size from size (if needed)
            if pixel_size is None and size is not None:
                pixel_size = self._pixel_size_from_img_size(size)

            # Reproj path in case
            reproj_path = self._get_utm_band_path(band=band.name, pixel_size=pixel_size)

            # Manage the case if we got a LAT LON product
            if not dst_crs.is_projected:
                if not reproj_path.is_file():
                    reproj_path = self._get_utm_band_path(
                        band=band.name, pixel_size=pixel_size, writable=True
                    )
                    # Warp band if needed
                    self._warp_band(
                        band_path,
                        reproj_path=reproj_path,
                        pixel_size=pixel_size,
                    )

                # Read band
                LOGGER.debug(f"Reading warped {band.name}.")
                band_arr = utils.read(
                    reproj_path,
                    pixel_size=pixel_size,
                    size=size,
                    resampling=Resampling.bilinear,
                    indexes=[self.bands[band].id],
                    **kwargs,
                )

            # Manage the case if we open a simple band (EOReader processed bands)
            elif dst.count == 1:
                # Read band
                band_arr = utils.read(
                    band_path,
                    pixel_size=pixel_size,
                    size=size,
                    resampling=Resampling.bilinear,
                    **kwargs,
                )

            # Manage the case if we open a stack (native DIMAP bands)
            else:
                # Read band
                band_arr = utils.read(
                    band_path,
                    pixel_size=pixel_size,
                    size=size,
                    resampling=Resampling.bilinear,
                    indexes=[self.bands[band].id],
                    **kwargs,
                )

        # Pop useless long name
        if "long_name" in band_arr.attrs:
            band_arr.attrs.pop("long_name")

        return band_arr

    def _load_bands(
        self,
        bands: list,
        pixel_size: float = None,
        size: Union[list, tuple] = None,
        **kwargs,
    ) -> dict:
        """
        Load bands as numpy arrays with the same pixel size (and same metadata).

        Args:
            bands list: List of the wanted bands
            pixel_size (float): Band pixel size in meters
            size (Union[tuple, list]): Size of the array (width, height). Not used if pixel_size is provided.
            kwargs: Other arguments used to load bands
        Returns:
            dict: Dictionary {band_name, band_xarray}
        """
        # Return empty if no band are specified
        if not bands:
            return {}

        # Get band paths
        if pixel_size is None and size is not None:
            pixel_size = self._pixel_size_from_img_size(size)
        band_paths = self.get_band_paths(bands, pixel_size=pixel_size, **kwargs)

        # Open bands and get array (resampled if needed)
        band_arrays = self._open_bands(
            band_paths, pixel_size=pixel_size, size=size, **kwargs
        )

        return band_arrays

    def _manage_invalid_pixels(
        self, band_arr: xr.DataArray, band: BandNames, **kwargs
    ) -> xr.DataArray:
        """
        Manage invalid pixels (Nodata, saturated, defective...)

        Args:
            band_arr (xr.DataArray): Band array
            band (BandNames): Band name as an SpectralBandNames
            kwargs: Other arguments used to load bands

        Returns:
            xr.DataArray: Cleaned band array
        """
        return self._manage_nodata(band_arr, band, **kwargs)

    def _manage_nodata(
        self, band_arr: xr.DataArray, band: BandNames, **kwargs
    ) -> xr.DataArray:
        """
        Manage only nodata pixels

        Args:
            band_arr (xr.DataArray): Band array
            band (BandNames): Band name as an SpectralBandNames
            kwargs: Other arguments used to load bands

        Returns:
            xr.DataArray: Cleaned band array
        """
        # If nodata not set, set it here
        if not band_arr.rio.encoded_nodata:
            band_arr = rasters.set_nodata(band_arr, self._raw_nodata)

        return band_arr

    def _get_condensed_name(self) -> str:
        """
        Get PlanetScope products condensed name ({date}_PLD_{product_type}_{band_combi}).

        Returns:
            str: Condensed name
        """
        return f"{self.get_datetime()}_{self.constellation.name}_{self.product_type.name}_{self.band_combi.name}_{self._job_id}"

    def _get_path(self, filename: str = "", extension: str = "") -> AnyPathType:
        """
        Get either the archived path of the normal path of an asset

        Args:
            filename (str): Filename with wildcards
            extension (str): Extension

        Returns:
            Union[list, AnyPathType]: Path or list of paths (needs this because of potential mosaic)

        """
        prod_path = []
        try:
            if filename and not filename.startswith("*"):
                filename = f"*{filename}"

            if self.is_archived:
                prod_path = path.get_archived_rio_path(
                    self.path,
                    rf".{filename}.*\.{extension}",
                )
            else:
                prod_path = next(self.path.glob(f"{filename}*.{extension}"))

        except (FileNotFoundError, IndexError):
            LOGGER.warning(
                f"No file corresponding to *{filename}*.{extension} found in {self.path}"
            )

        return prod_path

    def _get_utm_band_path(
        self,
        band: str,
        pixel_size: Union[float, tuple, list] = None,
        writable: bool = False,
    ) -> AnyPathType:
        """
        Create the UTM band path

        Args:
            band (str): Band in string as written on the filepath
            pixel_size (Union[float, tuple, list]): Pixel size of the wanted UTM band
            writable (bool): Do we need to write the UTM band ?

        Returns:
            AnyPathType: UTM band path
        """
        res_str = self._pixel_size_to_str(pixel_size)

        return self._get_band_folder(writable).joinpath(
            f"{self.condensed_name}_{band}_{res_str}.vrt"
        )

    def _warp_band(
        self,
        band_path: AnyPathStrType,
        reproj_path: AnyPathStrType,
        pixel_size: float = None,
    ) -> None:
        """
        Warp band to UTM

        Args:
            band_path (AnyPathStrType): Band path to warp
            reproj_path (AnyPathStrType): Path where to write the reprojected band
            pixel_size (int): Band pixel size in meters

        """
        # Do not warp if existing file
        if reproj_path.is_file():
            return

        if not pixel_size:
            pixel_size = self.pixel_size

        LOGGER.info(
            f"Warping {path.get_filename(band_path)} to UTM with a {pixel_size} m pixel size."
        )

        # Read band
        with rasterio.open(str(band_path)) as src:
            # Calculate transform
            utm_tr, utm_w, utm_h = warp.calculate_default_transform(
                src.crs,
                self.crs(),
                src.width,
                src.height,
                *src.bounds,
                resolution=pixel_size,
            )

            vrt_options = {
                "crs": self.crs(),
                "transform": utm_tr,
                "height": utm_h,
                "width": utm_w,
                # TODO: go nearest to speed up results ?
                "resampling": Resampling.bilinear,
                "nodata": self._raw_nodata,
                # Float32 is the max possible
                "warp_mem_limit": 32
                * int(os.getenv(TILE_SIZE, DEFAULT_TILE_SIZE)) ** 2
                / 1e6,
                "dtype": src.meta["dtype"],
                "num_threads": MAX_CORES,
            }
            with rasterio.Env(
                **{"GDAL_NUM_THREADS": "ALL_CPUS", "NUM_THREADS": "ALL_CPUS"}
            ):
                with WarpedVRT(src, **vrt_options) as vrt:
                    rio_shutil.copy(vrt, reproj_path, driver="vrt")

    def _get_default_utm_band(
        self, pixel_size: float = None, size: Union[list, tuple] = None
    ) -> AnyPathType:
        """
        Get the default UTM band:
        - If one already exists, get it
        - If not, create reproject (if needed) the GREEN band

        Args:
            pixel_size (int): Band pixel size in meters
            size (Union[tuple, list]): Size of the array (width, height). Not used if pixel_size is provided.

        Returns:
            str: Default UTM path
        """
        # Manage pixel_size
        if pixel_size is None and size is not None:
            pixel_size = self._pixel_size_from_img_size(size)
        def_res = pixel_size if pixel_size else self.pixel_size

        # Get default band path
        default_band = self.get_default_band()
        def_path = self.get_band_paths([default_band], pixel_size=def_res)[default_band]

        # First look for reprojected bands
        res_str = self._pixel_size_to_str(pixel_size)
        warped_regex = f"*{self.condensed_name}_*_{res_str}.tif"
        reproj_bands = list(self._get_band_folder().glob(warped_regex))

        if len(reproj_bands) == 0:
            # Check in the writeable band folder
            reproj_bands = list(self._get_band_folder(writable=True).glob(warped_regex))

            if len(reproj_bands) == 0:
                # Manage the case if we got a LAT LON product
                with rasterio.open(str(def_path)) as dst:
                    dst_crs = dst.crs

                if not dst_crs.is_projected:
                    def_band = self.get_default_band()
                    utm_path = self._get_utm_band_path(
                        band=def_band.name, pixel_size=pixel_size
                    )

                    # Warp band if needed
                    if not utm_path.is_file():
                        utm_path = self._get_utm_band_path(
                            band=def_band.name, pixel_size=pixel_size, writable=True
                        )
                        self._warp_band(
                            def_path,
                            reproj_path=utm_path,
                            pixel_size=pixel_size,
                        )
                else:
                    utm_path = def_path
            else:
                utm_path = AnyPath(reproj_bands[0])
        else:
            utm_path = AnyPath(reproj_bands[0])

        return utm_path

    def default_transform(self, **kwargs) -> (affine.Affine, int, int, CRS):
        """
        Returns default transform data of the default band (UTM),
        as the :code:`rasterio.warp.calculate_default_transform` does:
        - transform
        - width
        - height
        - CRS

        Overload in order not to reproject WGS84 data

        Args:
            kwargs: Additional arguments

        Returns:
            Affine, int, int, CRS: transform, width, height, CRS

        """
        default_band = self.get_default_band()
        def_path = self.get_band_paths(
            [default_band], pixel_size=self.pixel_size, **kwargs
        )[default_band]
        with rasterio.open(str(def_path)) as dst:
            return dst.transform, dst.width, dst.height, dst.crs

    @abstractmethod
    def _get_tile_path(self) -> AnyPathType:
        """
        Get the VHR tile path

        Returns:
            AnyPathType: VHR filepath
        """
        raise NotImplementedError

    @abstractmethod
    def _get_job_id(self) -> AnyPathType:
        """
        Get VHR job ID

        Returns:
            AnyPathType: VHR product ID
        """
        raise NotImplementedError

    def get_raw_band_paths(self, **kwargs) -> dict:
        """
        Return the raw band paths.

        Args:
            kwargs: Additional arguments

        Returns:
            dict: Dictionary containing the path of each queried band
        """
        raw_band_paths = {}
        for band in self.get_existing_bands():
            raw_band_paths[band] = self._get_tile_path()
        return raw_band_paths



# -*- coding: utf-8 -*-
# Copyright 2024, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of eoreader project
#     https://github.com/sertit/eoreader
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
DIMAP V1.1 products.

- VISION-1
- GEOSAT-2 (ex DEIMOS-2)
"""
import logging
import time
from abc import abstractmethod
from datetime import date, datetime
from typing import Union

import geopandas as gpd
import numpy as np
import xarray as xr
from rasterio import crs as riocrs
from sertit import geometry, rasters, vectors
from shapely.geometry import Polygon, box

from eoreader import DATETIME_FMT, EOREADER_NAME, cache
from eoreader.bands import BandNames
from eoreader.exceptions import InvalidProductError
from eoreader.products import VhrProduct
from eoreader.utils import simplify

LOGGER = logging.getLogger(EOREADER_NAME)


class DimapV1Product(VhrProduct):
    """
    Super Class of DIMAP V1.1 products.
    """

    @cache
    def crs(self) -> riocrs.CRS:
        """
        Get UTM projection of the tile

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"IMG_PHR1B_PMS_001"
            >>> prod = Reader().open(path)
            >>> prod.crs()
            CRS.from_epsg(32618)

        Returns:
            rasterio.crs.CRS: CRS object
        """
        raw_crs = self._get_raw_crs()

        if raw_crs.is_projected:
            utm = raw_crs
        else:
            # Open metadata
            root, _ = self.read_mtd()

            # Open the Bounding_Polygon
            vertices = list(root.iterfind(".//Dataset_Frame/Vertex"))

            # Get the mean lon lat
            lon = float(np.mean([float(v.findtext("FRAME_LON")) for v in vertices]))
            lat = float(np.mean([float(v.findtext("FRAME_LAT")) for v in vertices]))

            # Compute UTM crs from center long/lat
            utm = vectors.to_utm_crs(lon, lat)

        return utm

    @abstractmethod
    def _get_raw_crs(self) -> riocrs.CRS:
        """
        Get raw CRS of the tile

        Returns:
            rasterio.crs.CRS: CRS object
        """
        # Open metadata
        raise NotImplementedError

    @cache
    def extent(self, **kwargs) -> gpd.GeoDataFrame:
        """
        Get UTM extent of the tile.

        Returns:
            gpd.GeoDataFrame: Extent in UTM
        """
        # Get MTD XML file
        root, _ = self.read_mtd()

        # Compute extent corners
        corners = [
            [float(vertex.findtext("FRAME_LON")), float(vertex.findtext("FRAME_LAT"))]
            for vertex in root.iterfind(".//Dataset_Frame/Vertex")
        ]

        # When PRJ, Dataset_Frame is the footprint
        ds_frame = gpd.GeoDataFrame(
            geometry=[Polygon(corners)],
            crs=vectors.WGS84,
        ).to_crs(self.crs())

        extent = gpd.GeoDataFrame(
            geometry=[box(*ds_frame.total_bounds)],
            crs=self.crs(),
        )

        return extent

    @cache
    @simplify
    def footprint(self) -> gpd.GeoDataFrame:
        """
        Get real footprint in UTM of the products (without nodata, in french == emprise utile)

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"IMG_PHR1B_PMS_001"
            >>> prod = Reader().open(path)
            >>> prod.footprint()
                                                         gml_id  ...                                           geometry
            0  source_image_footprint-DS_PHR1A_20200511023124...  ...  POLYGON ((707025.261 9688613.833, 707043.276 9...
            [1 rows x 3 columns]

        Returns:
            gpd.GeoDataFrame: Footprint as a GeoDataFrame
        """
        # If ortho -> nodata is not set !
        if self.is_ortho:
            # Get footprint of the first band of the stack
            footprint_dezoom = 10
            arr = rasters.read(
                self.get_default_band_path(),
                resolution=self.pixel_size * footprint_dezoom,
                indexes=[1],
            )

            # Just in case
            arr = arr.fillna(0)

            # Vectorize the nodata band
            footprint = rasters.vectorize(
                arr, values=0, keep_values=False, dissolve=True
            )
            footprint = geometry.get_wider_exterior(footprint)
        else:
            # If not ortho -> default band has been orthorectified and nodata will be set
            footprint = rasters.get_footprint(self.get_default_band_path())

        return footprint.to_crs(self.crs())

    def get_datetime(self, as_datetime: bool = False) -> Union[str, datetime]:
        """
        Get the product's acquisition datetime, with format :code:`YYYYMMDDTHHMMSS` <-> :code:`%Y%m%dT%H%M%S`

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"IMG_PHR1B_PMS_001"
            >>> prod = Reader().open(path)
            >>> prod.get_datetime(as_datetime=True)
            datetime.datetime(2020, 5, 11, 2, 31, 58)
            >>> prod.get_datetime(as_datetime=False)
            '20200511T023158'

        Args:
            as_datetime (bool): Return the date as a datetime.datetime. If false, returns a string.

        Returns:
             Union[str, datetime.datetime]: Its acquisition datetime
        """
        if self.datetime is None:
            # Get MTD XML file
            root, _ = self.read_mtd()

            # GeoSAT-2
            date_str = root.findtext(".//START_TIME")

            if date_str:
                date_str = datetime.strptime(date_str, "%Y-%m-%dT%H:%M:%S").strftime(
                    DATETIME_FMT
                )
            else:
                # Vision-1
                date_str = root.findtext(".//IMAGING_DATE")
                time_str = root.findtext(".//IMAGING_TIME")
                if not date_str or not time_str:
                    raise InvalidProductError(
                        "Cannot find the product imaging date and time in the metadata file."
                    )

                # Convert to datetime
                date_dt = date.fromisoformat(date_str)
                try:
                    time_dt = time.strptime(time_str, "%H:%M:%S.%fZ")
                except ValueError:
                    # Sometimes without a Z
                    try:
                        time_dt = time.strptime(time_str, "%H:%M:%S.%f")
                    except ValueError:
                        # Sometimes without MICROSECONDS
                        time_dt = time.strptime(time_str, "%H:%M:%S")

                date_str = (
                    f"{date_dt.strftime('%Y%m%d')}T{time.strftime('%H%M%S', time_dt)}"
                )

            if as_datetime:
                date_str = datetime.strptime(date_str, DATETIME_FMT)

        else:
            date_str = self.datetime
            if not as_datetime:
                date_str = date_str.strftime(DATETIME_FMT)

        return date_str

    def _get_name_constellation_specific(self) -> str:
        """
        Set product real name from metadata

        Returns:
            str: True name of the product (from metadata)
        """
        # Get MTD XML file
        root, _ = self.read_mtd()

        # Open identifier
        name = root.findtext(".//DATASET_NAME")
        if not name:
            raise InvalidProductError("DATASET_NAME not found in metadata!")

        return name

    @cache
    def get_mean_sun_angles(self) -> (float, float):
        """
        Get Mean Sun angles (Azimuth and Zenith angles)

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"IMG_PHR1A_PMS_001"
            >>> prod = Reader().open(path)
            >>> prod.get_mean_sun_angles()
            (45.6624568841367, 30.219881316357643)

        Returns:
            (float, float): Mean Azimuth and Zenith angle
        """
        # Get MTD XML file
        root, _ = self.read_mtd()

        # Open zenith and azimuth angle
        elev_angle = float(root.findtext(".//SUN_ELEVATION"))
        azimuth_angle = float(root.findtext(".//SUN_AZIMUTH"))

        # From elevation to zenith
        zenith_angle = 90.0 - elev_angle

        return azimuth_angle, zenith_angle

    @cache
    def get_mean_viewing_angles(self) -> (float, float, float):
        """
        Get Mean Viewing angles (azimuth, off-nadir and incidence angles)

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.get_mean_viewing_angles()

        Returns:
            (float, float, float): Mean azimuth, off-nadir and incidence angles
        """
        # Get MTD XML file
        root, _ = self.read_mtd()

        # Open sensor azimuth, off nadir and incidence angles
        try:
            az = None
            for a in root.iterfind(".//Quality_Parameter"):
                if a.findtext("QUALITY_PARAMETER_CODE") == "SPACEMETRIC:SENSOR_AZIMUTH":
                    az = float(a.findtext("QUALITY_PARAMETER_VALUE"))
                    break

            incidence_angle = 90 - float(root.findtext(".//INCIDENCE_ANGLE"))
            off_nadir = float(root.findtext(".//VIEWING_ANGLE"))
        except TypeError:
            raise InvalidProductError(
                "SPACEMETRIC:SENSOR_AZIMUTH, INCIDENCE_ANGLE or VIEWING_ANGLE not found in metadata!"
            )

        return az, off_nadir, incidence_angle

    def _has_cloud_band(self, band: BandNames) -> bool:
        """
        Does this product has the specified cloud band ?
        """
        return False

    def _open_clouds(
        self,
        bands: list,
        pixel_size: float = None,
        size: Union[list, tuple] = None,
        **kwargs,
    ) -> dict:
        """
        Load cloud files as xarrays.

        Args:
            bands (list): List of the wanted bands
            pixel_size (int): Band pixel size in meters
            size (Union[tuple, list]): Size of the array (width, height). Not used if pixel_size is provided.
            kwargs: Additional arguments
        Returns:
            dict: Dictionary {band_name, band_xarray}
        """
        return {}

    def _toa_rad_to_toa_refl(
        self, rad_arr: xr.DataArray, band: BandNames, e0: float, dt: float = None
    ) -> xr.DataArray:
        """
        Compute TOA reflectance from TOA radiance

        See
        `here <https://www.intelligence-airbusds.com/automne/api/docs/v1.0/document/download/ZG9jdXRoZXF1ZS1kb2N1bWVudC02ODMwNQ==/ZG9jdXRoZXF1ZS1maWxlLTY4MzAy/vision-1-imagery-user-guide-20210217>`_
        (3.2.2) for more information.

        WARNING: in this formula, d**2 = 1 / sqrt(dt) !

        Args:
            rad_arr (xr.DataArray): TOA Radiance array
            band (BandNames): Band
            e0 (float): Solar spectral irradiance for the current band

        Returns:
            xr.DataArray: TOA Reflectance array
        """
        # Compute the coefficient converting TOA radiance in TOA reflectance
        if not dt:
            dt = self._sun_earth_distance_variation()
        _, sun_zen = self.get_mean_sun_angles()
        rad_sun_zen = np.deg2rad(sun_zen)
        toa_refl_coeff = np.pi / (e0 * dt * np.cos(rad_sun_zen))

        # LOGGER.debug(f"rad to refl coeff = {toa_refl_coeff}")
        return rad_arr.copy(data=toa_refl_coeff * rad_arr)

    def _get_job_id(self) -> str:
        """
        Get VHR job ID

        Returns:
            str: VHR product ID
        """
        return self.split_name[-2]



# -*- coding: utf-8 -*-
# Copyright 2024, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of eoreader project
#     https://github.com/sertit/eoreader
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Capella products.
Take a look
`here <https://support.capellaspace.com/hc/en-us/categories/360002612692-SAR-Imagery-Products>`_.
"""
import logging
from datetime import datetime
from enum import unique
from typing import Union

import geopandas as gpd
from affine import Affine
from dicttoxml import dicttoxml
from lxml import etree
from rasterio import CRS, transform
from sertit import files, path, vectors
from sertit.misc import ListEnum
from sertit.types import AnyPathStrType
from sertit.vectors import WGS84
from shapely.geometry import Point, box

from eoreader import DATETIME_FMT, EOREADER_NAME, cache
from eoreader.bands import SarBandNames as sab
from eoreader.exceptions import InvalidProductError
from eoreader.products import SarProduct, SarProductType
from eoreader.products.product import OrbitDirection

LOGGER = logging.getLogger(EOREADER_NAME)


@unique
class CapellaProductType(ListEnum):
    """
    Capella products types.
    Take a look
    `here <https://support.capellaspace.com/hc/en-us/articles/360039702691-SAR-Data-Formats>`_.
    """

    SLC = "SLC"
    """
    Single Look Complex (SLC)

    - Contains both amplitude and phase of the radar signal
    - Range-compressed and focused SAR image in slant-range geometry
    - Georeferenced using orbit data and Range-Doppler projected
    """

    GEC = "GEC"
    """
    Geocoded Ellipsoid Corrected (GEC)

    - Contains amplitude information only
    - Range-compressed, detected, focused and multi-looked SAR image
    - Multi-look techniques applied to enhance radiometric resolution
    - Resampled and projected onto WGS84 ellipsoid with average scene center height
    - Universal Transverse Mercator (UTM) and Universal Polar Stereographic (UPS) projections
    """

    GEO = "GEO"
    """
    Geocoded Terrain Corrected (GEO)

    - Contains amplitude information only
    - Range-compressed, detected, focused and multi-looked SAR image
    - Multi-look techniques applied to enhance radiometric resolution
    - Terrain-height corrected using a high-resolution Digital Elevation Model (DEM)
    - Universal Transverse Mercator (UTM) and Universal Polar Stereographic (UPS) projections
    """

    SICD = "SICD"
    """
    Sensor Independent Complex Data (SICD)

    - Contains both amplitude and phase of the radar signal
    - Range-compressed and focused SAR image in slant-range geometry
    - Sensor independent format

    Not used by EOReader.
    """

    SIDD = "SIDD"
    """
    Sensor Independent Derived Data (SIDD)

    - Contains amplitude information only
    - Range-compressed, detected, focused and multi-looked SAR image
    - Multi-look techniques applied to enhance radiometric resolution
    - Planar Gridded Display (PGD) projection
    - Sensor independent format

    Not used by EOReader.
    """

    CPHD = "CPHD"
    """
    Compensated Phase History Data (CPHD)

    - Contains raw phase history data that is compensated for hardware timing & platform motion
    - Sensor independent format
    - Only available to United States Government customers

    Not used by EOReader.
    """


@unique
class CapellaSensorMode(ListEnum):
    """
    Capella imaging mode.
    Take a look
    `here <https://support.capellaspace.com/hc/en-us/articles/360059224291-What-SAR-imagery-products-are-available-with-Capella->`_.
    """

    SM = "stripmap"
    """Stripmap"""

    SP = "spotlight"
    """Spotlight"""

    SS = "sliding_spotlight"
    """Sliding Spotlight"""


class CapellaProduct(SarProduct):
    """
    Class for Capella Products
    Take a look
    `here <https://support.capellaspace.com/hc/en-us/categories/360002612692-SAR-Imagery-Products>`_.
    """

    def __init__(
        self,
        product_path: AnyPathStrType,
        archive_path: AnyPathStrType = None,
        output_path: AnyPathStrType = None,
        remove_tmp: bool = False,
        **kwargs,
    ) -> None:
        self._has_stac_mtd = False

        # Initialization from the super class
        super().__init__(product_path, archive_path, output_path, remove_tmp, **kwargs)

    def _set_pixel_size(self) -> None:
        """
        Set product default pixel size (in meters)
        See here
        `here <https://support.capellaspace.com/hc/en-us/articles/360059224291-What-SAR-imagery-products-are-available-with-Capella->`_
        """
        # Using az resolution
        if self.sensor_mode == CapellaSensorMode.SP:
            def_pixel_size = 0.35
            def_res = 0.5
        elif self.sensor_mode == CapellaSensorMode.SM:
            def_pixel_size = 0.6
            def_res = 1.0
        elif self.sensor_mode == CapellaSensorMode.SS:
            def_pixel_size = 0.8
            def_res = 1.2
        else:
            raise InvalidProductError(f"Unknown sensor mode: {self.sensor_mode}")
        self.pixel_size = def_pixel_size
        self.resolution = def_res

    def _set_instrument(self) -> None:
        """
        Set instrument
        """
        self.instrument = "SAR X-band"

    def _pre_init(self, **kwargs) -> None:
        """
        Function used to pre_init the products
        (setting needs_extraction and so on)
        """
        self._band_folder = self.path

        # SNAP cannot process its archive
        self.needs_extraction = True

        # Its original filename is its name
        self._use_filename = True

        # Pre init done by the super class
        super()._pre_init(**kwargs)

    def _post_init(self, **kwargs) -> None:
        """
        Function used to post_init the products
        (setting product-type, band names and so on)
        """
        # Private attributes
        self.snap_filename = str(next(self.path.glob("*CAPELLA*.json")).name)
        try:
            self._raw_band_regex = str(next(self.path.glob(f"{self.name}.tif")).name)
        except StopIteration:
            # For SICD and SIDD
            self._raw_band_regex = str(next(self.path.glob(f"{self.name}.ntf")).name)

        # Post init done by the super class
        super()._post_init(**kwargs)

    @cache
    def wgs84_extent(self) -> gpd.GeoDataFrame:
        """
        Get the WGS84 extent of the file before any reprojection.
        This is useful when the SAR pre-process has not been done yet.

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"1011117-766193"
            >>> prod = Reader().open(path)
            >>> prod.wgs84_extent()
                                                        geometry
            0  POLYGON ((108.09797 15.61011, 108.48224 15.678...

        Returns:
            gpd.GeoDataFrame: WGS84 extent as a gpd.GeoDataFrame
        """
        if self._has_stac_mtd:
            mtd_file = next(self.path.glob(f"{self.name}.json"))
            extent = vectors.read(mtd_file)
        else:
            extent = None
            root, _ = self.read_mtd()

            # Get image size
            height = int(root.findtext(".//rows"))
            width = int(root.findtext(".//columns"))

            # Investigate image geometry
            img_geom = root.find(".//image_geometry")
            geom_type = img_geom.findtext("type")

            # Use given geotransform if existing
            if geom_type == "geotransform":
                tf = [
                    float(it.text)
                    for it in img_geom.find("geotransform").iterfind("item")
                ]
                # TODO: manage not WKT case
                crs = img_geom.find("coordinate_system").findtext("wkt")
                if crs:
                    # Convert to rasterio
                    west, south, east, north = transform.array_bounds(
                        height, width, transform=Affine.from_gdal(*tf)
                    )
                    extent = gpd.GeoDataFrame(
                        geometry=[box(minx=west, miny=north, maxx=east, maxy=south)],
                        crs=CRS.from_string(crs),
                    )

            # Use center pixel
            if extent is None:
                # Get center pixel point in UTM
                center_pixel = root.find(".//center_pixel")
                target_position = [
                    float(it.text)
                    for it in center_pixel.find("target_position").iterfind("item")
                ]
                center_pix = gpd.GeoDataFrame(
                    geometry=[Point(target_position)],
                    crs={"proj": "geocent", "ellps": "WGS84", "datum": "WGS84"},
                ).to_crs(WGS84)
                center_pix.to_crs(center_pix.extent_wgs84.estimate_utm_crs())

                # Get pixel spacing in meters
                pixel_spacing_h = float(root.findtext(".//pixel_spacing_row"))
                pixel_spacing_w = float(root.findtext(".//pixel_spacing_column"))

                # Compute offset from center of image
                offset_h = pixel_spacing_h * height / 2
                offset_w = pixel_spacing_w * width / 2

                tl_corner = center_pix.translate(xoff=-offset_w, yoff=-offset_h)
                br_corner = center_pix.translate(xoff=offset_w, yoff=offset_h)

                extent = gpd.GeoDataFrame(
                    geometry=[
                        box(
                            minx=tl_corner.x.iat[0],
                            miny=tl_corner.y.iat[0],
                            maxx=br_corner.x.iat[0],
                            maxy=br_corner.y.iat[0],
                        )
                    ],
                    crs=CRS.from_string(self.crs()),
                )

        return extent.to_crs(WGS84)

    def _set_product_type(self) -> None:
        """Set products type"""
        # Open identifier
        prod_type = self.split_name[3]
        self.product_type = getattr(CapellaProductType, prod_type)

        if self.product_type in [CapellaProductType.GEO, CapellaProductType.GEC]:
            self.sar_prod_type = SarProductType.GDRG
        elif self.product_type == CapellaProductType.SLC:
            self.sar_prod_type = SarProductType.CPLX
        else:
            raise NotImplementedError(
                f"{self.product_type.value} product type is not handled by EOReader"
            )

    def _set_sensor_mode(self) -> None:
        """Get sensor mode"""
        sensor_mode = self.split_name[2]
        self.sensor_mode = getattr(CapellaSensorMode, sensor_mode)

    def get_datetime(self, as_datetime: bool = False) -> Union[str, datetime]:
        """
        Get the product's acquisition datetime, with format :code:`YYYYMMDDTHHMMSS` <-> :code:`%Y%m%dT%H%M%S`

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"1011117-766193"
            >>> prod = Reader().open(path)
            >>> prod.get_datetime(as_datetime=True)
            datetime.datetime(2020, 10, 28, 22, 46, 25)
            >>> prod.get_datetime(as_datetime=False)
            '20201028T224625'

        Args:
            as_datetime (bool): Return the date as a datetime.datetime. If false, returns a string.

        Returns:
             Union[str, datetime.datetime]: Its acquisition datetime
        """
        if self.datetime is None:
            # Get MTD XML file
            root, _ = self.read_mtd()

            # Open identifier
            if self._has_stac_mtd:
                acq_date = root.findtext(".//datetime")
            else:
                acq_date = root.findtext(".//start_timestamp")
            if not acq_date:
                raise InvalidProductError(
                    "datetime or start_timestamp not found in metadata!"
                )

            # Convert to datetime (too many microseconds)
            date = datetime.strptime(acq_date.split(".")[0], "%Y-%m-%dT%H:%M:%S")
        else:
            date = self.datetime

        if not as_datetime:
            date = date.strftime(DATETIME_FMT)

        return date

    def _get_name_constellation_specific(self) -> str:
        """
        Set product real name from metadata

        Returns:
            str: True name of the product (from metadata)
        """
        name = None
        for file in self.path.glob("*.tif"):
            if "preview" not in file.name:
                name = path.get_filename(file)

        return name

    @cache
    def _read_mtd(self) -> (etree._Element, dict):
        """
        Read GeoJSON metadata and outputs its as a metadata XML root and its namespaces as an empty dict

        Returns:
            (etree._Element, dict): Metadata XML root and its namespaces as a dict
        """

        # MTD are JSON
        try:
            try:
                mtd_file = next(self.path.glob(f"{self.name}.json"))
                self._has_stac_mtd = True
            except StopIteration:
                try:
                    LOGGER.warning(
                        f"Non available STAC metadata for the product {self.name}. Opening Extended Metadata instead."
                    )
                    mtd_file = next(self.path.glob(f"{self.name}*.json"))
                    self._has_stac_mtd = False
                except StopIteration as ex:
                    raise InvalidProductError(
                        f"Metadata file not found in {self.path}"
                    ) from ex

            data = files.read_json(mtd_file, print_file=False)
            root = etree.fromstring(dicttoxml(data))
        except etree.XMLSyntaxError:
            raise InvalidProductError(
                f"Cannot convert metadata to XML for {self.path}!"
            )

        return root, {}

    def get_raw_band_paths(self, **kwargs) -> dict:
        """
        Return the existing path of the VV band (as they come with the archived products).
        Capella product only contains a VV band !

        Args:
            **kwargs: Additional arguments

        Returns:
            dict: Dictionary containing the path of every band existing in the raw products
        """
        band_paths = {}
        try:
            pol = sab.from_value(self.split_name[4])
            band_paths[pol] = path.get_file_in_dir(
                self._band_folder, self._raw_band_regex, exact_name=True, get_list=False
            )
        except FileNotFoundError:
            raise InvalidProductError(
                f"An {self.constellation.name} product should at least contain one band !"
            )

        return band_paths

    def get_quicklook_path(self) -> str:
        """
        Get quicklook path if existing.

        Returns:
            str: Quicklook path
        """
        quicklook_path = None
        try:
            quicklook_path = str(next(self.path.glob("*.png")))
        except StopIteration:
            LOGGER.warning(f"No quicklook found in {self.condensed_name}")

        return quicklook_path

    # @cache
    def get_orbit_direction(self) -> OrbitDirection:
        """
        Get cloud cover as given in the metadata

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.get_orbit_direction().value
            "DESCENDING"

        Returns:
            OrbitDirection: Orbit direction (ASCENDING/DESCENDING)
        """
        ob = None
        if self._has_stac_mtd:
            root, _ = self.read_mtd()

            ob = root.findtext(".//key[@name='sat:orbit_state']")
            ob = OrbitDirection.from_value(ob.upper())

        if ob is None:
            ob = super().get_orbit_direction()

        return ob



# -*- coding: utf-8 -*-
# Copyright 2024, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of eoreader project
#     https://github.com/sertit/eoreader
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
PlanetScope products.
See
`Product specs <https://assets.planet.com/docs/Planet_Combined_Imagery_Product_Specs_letter_screen.pdf>`_
and `Planet documentation <https://developers.planet.com/docs/data/planetscope/>`_
for more information.
"""
import logging
from collections import defaultdict
from datetime import datetime
from enum import unique
from typing import Union

import numpy as np
import rasterio
import xarray as xr
from lxml import etree
from sertit import path, rasters, xml
from sertit.misc import ListEnum
from sertit.types import AnyPathType
from sertit.vectors import WGS84

from eoreader import DATETIME_FMT, EOREADER_NAME
from eoreader.bands import (
    BLUE,
    CA,
    GREEN,
    GREEN_1,
    NARROW_NIR,
    NIR,
    RED,
    VRE_1,
    VRE_2,
    VRE_3,
    YELLOW,
    BandNames,
    SpectralBand,
)
from eoreader.exceptions import InvalidProductError
from eoreader.products.optical.planet_product import PlanetProduct
from eoreader.stac import GSD, ID, NAME, WV_MAX, WV_MIN

LOGGER = logging.getLogger(EOREADER_NAME)


@unique
class PlaInstrument(ListEnum):
    """PlanetScope instrument
    See `Planet documentation <https://developers.planet.com/docs/apis/data/sensors/>`__
    for more information.
    """

    PS2 = "Dove Classic (PS2)"
    """
    Dove Classic (PS2) Instrument: Four-band frame Image with a split-frame VIS+NIR filter
    """

    PS2_SD = "Dove-R (PS2.SD)"
    """
    Dove-R (PS2.SD) Instrument:
    Four-band frame imager with butcher-block filter providing blue, green, red,and NIR stripes
    """

    PSB_SD = "SuperDove (PSB.SD)"
    """
    SuperDove (PSB.SD) Instrument:
    Eight-band frame imager with butcher-block filter providing:

    - coastal blue,
    - blue,
    - green I,
    - green II,
    - yellow,
    - red,
    - red-edge,
    - NIR stripes
    """


@unique
class PlaProductType(ListEnum):
    """PlanetScope product types (processing levels)"""

    L1B = "Basic Scene Product"
    """
    **PlanetScope Basic Scene Product (Level 1B)**

    Scaled Top of Atmosphere Radiance(at sensor) and sensor corrected product.
    This product has scene based framing and is not projected to a cartographic projection.
    Radiometric and sensor corrections are applied to the data.
    """

    L3B = "Ortho Scene Product"
    """
    **PlanetScope Ortho Scene Product (Level 3B)**

    Orthorectified, scaled Top of Atmosphere Radiance (at sensor) or Surface Reflectance image product
    suitable for analytic and visual applications.
    This product has scene based framing and projected to a cartographic projection.

    **PSScene3Band**

    PlanetScope 3-band multispectral basic and orthorectified scenes.
    This data set includes imagery from PlanetScope-0 and PlanetScope-1 sensors
    as well as full-frame and split-frame PlanetScope-2 sensors.
    Newer PSScene3Band items have a corresponding PSScene4Band item.

    Resampled to 3.0m.

    **PSScene4Band**

    PlanetScope 4-band multispectral basic and orthorectified scenes.
    This data set includes imagery from all PlanetScope sensors.
    All PSScene4Band items have a corresponding PSScene3Band item.

    Resampled to 3.0m.
    """
    """
    **PSScene (Not found anywhere else)**

    PlanetScope 8-band multispectral basic and orthorectified scenes.
    This data set includes imagery from all PlanetScope sensors.

    Naming: <acq date>_<acq time>_<acq time seconds ms>_<satellite_id>_<productLevel>_<bandProduct>.<ext>

    Asset Types:
    ortho_analytic_4b       Radiometrically-calibrated analytic image stored as 16-bit scaled radiance.
    ortho_analytic_8b       Radiometrically-calibrated analytic image stored as 16-bit scaled radiance.
    ortho_analytic_8b_sr    PlanetScope atmospherically corrected surface reflectance product.
    ortho_analytic_8b_xml   Radiometrically-calibrated analytic image metadata.
    ortho_analytic_4b_sr    PlanetScope atmospherically corrected surface reflectance product.
    ortho_analytic_4b_xml   Radiometrically-calibrated analytic image metadata.
    basic_analytic_4b       Unorthorectified radiometrically-calibrated analytic image stored as 16-bit scaled radiance.
    basic_analytic_8b       Unorthorectified radiometrically-calibrated analytic image stored as 16-bit scaled radiance.
    basic_analytic_8b_xml   Unorthorectified radiometrically-calibrated analytic image metadata
    basic_analytic_4b_rpc   RPC for unorthorectified analytic image stored as 12-bit digital numbers.
    basic_analytic_4b_xml   Unorthorectified radiometrically-calibrated analytic image metadata.
    basic_udm2              Unorthorectified usable data mask (Cloud 2.0)
    ortho_udm2              Usable data mask (Cloud 2.0)
    ortho_visual            Visual image with color-correction
    """

    L3A = "Ortho Tile Product"
    """
    **PlanetScope Ortho Tile Product (Level 3A)**

    Radiometric and sensor corrections applied to the data.
    Imagery is orthorectified and projected to a UTM projection.

    **PSOrthoTile**

    PlanetScope Ortho Tiles as 25 km x 25 km UTM tiles. This data set includes imagery from all PlanetScope sensors.
    Resampled to 3.125m.

    Naming: <strip_id>_<tile_id>_<acquisition date>_<satellite_id>_<bandProduct>.<extension>

    Product band order:

    - Band 1 = Blue
    - Band 2 = Green
    - Band 3 = Red
    - Band 4 = Near-infrared (analytic products only)

    Analytic 5B Product band order:

    - Band 1 = Blue
    - Band 2 = Green
    - Band 3 = Red
    - Band 4 = Red-Edge
    - Band 5 = Near-infrared
    """


class PlaProduct(PlanetProduct):
    """
    Class of PlanetScope products.
    See `here <https://assets.planet.com/docs/Planet_Combined_Imagery_Product_Specs_letter_screen.pdf>`__
    for more information.

    The scaling factor to retrieve the calibrated radiance is 0.01.
    """

    def _pre_init(self, **kwargs) -> None:
        """
        Function used to pre_init the products
        (setting needs_extraction and so on)
        """
        self.constellation = self._get_constellation()
        super()._pre_init(**kwargs)

    def _post_init(self, **kwargs) -> None:
        """
        Function used to post_init the products
        (setting sensor type, band names and so on)
        """
        self._has_cloud_cover = True

        # Post init done by the super class
        super()._post_init(**kwargs)

    def _set_pixel_size(self) -> None:
        """
        Set product default pixel size (in meters)
        """
        # Ortho Tiles
        if self.product_type == PlaProductType.L3A:
            self.pixel_size = 3.125
        # Ortho Scene
        else:
            self.pixel_size = 3.0

    def _set_instrument(self) -> None:
        """
        Set instrument
        """
        # Get MTD XML file
        root, nsmap = self.read_mtd()

        # Manage constellation
        instr_node = root.find(f".//{nsmap['eop']}Instrument")
        instrument = instr_node.findtext(f"{nsmap['eop']}shortName")

        if not instrument:
            raise InvalidProductError("Cannot find the Instrument in the metadata file")

        # Set correct constellation
        self.instrument = getattr(PlaInstrument, instrument.replace(".", "_"))

    def _get_spectral_bands(self) -> dict:
        """
        See `here <https://assets.planet.com/docs/Planet_Combined_Imagery_Product_Specs_letter_screen.pdf>`_ for more information.

        Returns:
            dict: PlanetScope spectral bands
        """
        gsd = 3.7

        blue = SpectralBand(
            eoreader_name=BLUE,
            **{NAME: "Blue", ID: 2, GSD: gsd, WV_MIN: 465, WV_MAX: 515},
        )

        green = SpectralBand(
            eoreader_name=GREEN,
            **{NAME: "Green", ID: 4, GSD: gsd, WV_MIN: 547, WV_MAX: 583},
        )

        red = SpectralBand(
            eoreader_name=RED,
            **{NAME: "Red", ID: 6, GSD: gsd, WV_MIN: 650, WV_MAX: 680},
        )

        nir = SpectralBand(
            eoreader_name=NIR,
            **{NAME: "NIR", ID: 8, GSD: gsd, WV_MIN: 845, WV_MAX: 885},
        )

        # Create spectral bands
        if self.instrument == PlaInstrument.PSB_SD:
            spectral_bands = {
                "ca": SpectralBand(
                    eoreader_name=CA,
                    **{NAME: "Coastal Blue", ID: 1, GSD: gsd, WV_MIN: 431, WV_MAX: 452},
                ),
                "blue": blue,
                "green1": SpectralBand(
                    eoreader_name=GREEN_1,
                    **{NAME: "Green I", ID: 3, GSD: gsd, WV_MIN: 513, WV_MAX: 549},
                ),
                "green": green.update(name="Green II"),
                "yellow": SpectralBand(
                    eoreader_name=YELLOW,
                    **{NAME: "Yellow", ID: 5, GSD: gsd, WV_MIN: 600, WV_MAX: 620},
                ),
                "red": red,
                "vre": SpectralBand(
                    eoreader_name=VRE_1,
                    **{NAME: "Red-Edge", ID: 7, GSD: gsd, WV_MIN: 697, WV_MAX: 713},
                ),
                "nir": nir,
            }
        elif self.instrument == PlaInstrument.PS2_SD:
            spectral_bands = {
                "blue": blue.update(**{ID: 1, WV_MIN: 464, WV_MAX: 517}),
                "green": green.update(**{ID: 2, WV_MIN: 547, WV_MAX: 585}),
                "red": red.update(**{ID: 3, WV_MIN: 650, WV_MAX: 682}),
                "nir": nir.update(**{ID: 4, WV_MIN: 846, WV_MAX: 888}),
            }
        elif self.instrument == PlaInstrument.PS2:
            spectral_bands = {
                "blue": blue.update(**{ID: 1, WV_MIN: 455, WV_MAX: 515}),
                "green": green.update(**{ID: 2, WV_MIN: 500, WV_MAX: 590}),
                "red": red.update(**{ID: 3, WV_MIN: 590, WV_MAX: 670}),
                "nir": nir.update(**{ID: 4, WV_MIN: 780, WV_MAX: 860}),
            }
        else:
            raise InvalidProductError(
                f"Non recognized PlanetScope Instrument: {self.instrument}"
            )

        return spectral_bands

    def _get_band_map(self, nof_bands, **kwargs) -> dict:
        """
        Get band map
        """
        # Open spectral bands
        ca = kwargs.get("ca")
        blue = kwargs.get("blue")
        green = kwargs.get("green")
        green1 = kwargs.get("green1")
        red = kwargs.get("red")
        nir = kwargs.get("nir")
        vre = kwargs.get("vre")
        yellow = kwargs.get("yellow")

        if nof_bands == 3:
            band_map = {
                BLUE: blue.update(id=1),
                GREEN: green.update(id=2),
                RED: red.update(id=3),
            }
        elif nof_bands == 4:
            band_map = {
                BLUE: blue.update(id=1),
                GREEN: green.update(id=2),
                RED: red.update(id=3),
                NIR: nir.update(id=4),
                NARROW_NIR: nir.update(id=4),
            }
        elif nof_bands == 5:
            band_map = {
                BLUE: blue.update(id=1),
                GREEN: green.update(id=2),
                RED: red.update(id=3),
                VRE_1: vre.update(id=4),
                VRE_2: vre.update(id=4),
                VRE_3: vre.update(id=4),
                NIR: nir.update(id=5),
                NARROW_NIR: nir.update(id=5),
            }
        elif nof_bands == 8:
            band_map = {
                CA: ca,
                BLUE: blue,
                GREEN_1: green1,
                GREEN: green,
                RED: red,
                YELLOW: yellow,
                VRE_1: vre,
                NIR: nir,
                NARROW_NIR: nir,
            }
        else:
            raise InvalidProductError(
                f"Unusual number of bands ({nof_bands}) for {self.path}. "
                f"Please check the validity of your product"
            )

        return band_map

    def _map_bands(self):
        """
        Map bands
        """
        # Get MTD XML file
        root, nsmap = self.read_mtd()

        # Manage bands of the product
        nof_bands = int(root.findtext(f".//{nsmap[self._nsmap_key]}numBands"))

        # Set the band map
        self.bands.map_bands(
            self._get_band_map(nof_bands, **self._get_spectral_bands())
        )

    def _set_product_type(self) -> None:
        """Set products type"""
        # Get MTD XML file
        root, nsmap = self.read_mtd()

        # Manage product type
        prod_type = root.findtext(f".//{nsmap['eop']}productType")
        if not prod_type:
            raise InvalidProductError(
                "Cannot find the product type in the metadata file"
            )

        # Set correct product type
        self.product_type = getattr(PlaProductType, prod_type)
        if self.product_type == PlaProductType.L1B:
            raise NotImplementedError(
                f"Basic Scene Product are not managed for Planet products {self.path}"
            )

    def get_datetime(self, as_datetime: bool = False) -> Union[str, datetime]:
        """
        Get the product's acquisition datetime, with format :code:`YYYYMMDDTHHMMSS` <-> :code:`%Y%m%dT%H%M%S`

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"SENTINEL2A_20190625-105728-756_L2A_T31UEQ_C_V2-2"
            >>> prod = Reader().open(path)
            >>> prod.get_datetime(as_datetime=True)
            datetime.datetime(2019, 6, 25, 10, 57, 28, 756000), fetched from metadata, so we have the ms
            >>> prod.get_datetime(as_datetime=False)
            '20190625T105728'

        Args:
            as_datetime (bool): Return the date as a datetime.datetime. If false, returns a string.

        Returns:
             Union[str, datetime.datetime]: Its acquisition datetime
        """
        if self.datetime is None:
            # Get MTD XML file
            root, nsmap = self.read_mtd()
            datetime_str = root.findtext(f".//{nsmap['eop']}acquisitionDate")
            if not datetime_str:
                raise InvalidProductError(
                    "Cannot find acquisitionDate in the metadata file."
                )

            # Convert to datetime
            datetime_str = datetime_str.split("+")[0]
            try:
                datetime_str = datetime.strptime(datetime_str, "%Y-%m-%dT%H:%M:%S")
            except ValueError:
                datetime_str = datetime.strptime(datetime_str, "%Y-%m-%dT%H:%M:%S.%f")

        else:
            datetime_str = self.datetime

        if not as_datetime:
            datetime_str = datetime_str.strftime(DATETIME_FMT)

        return datetime_str

    def _get_stack_path(self, as_list: bool = False) -> Union[str, list]:
        """
        Get Planet stack path(s)

        Args:
            as_list (bool): Get stack path as a list (useful if several subdatasets are present)

        Returns:
            Union[str, list]: Stack path(s)
        """
        if self._merged:
            stack_path, _ = self._get_out_path(f"{self.condensed_name}_analytic.vrt")
            if as_list:
                stack_path = [stack_path]
        else:
            stack_path = self._get_path(
                "Analytic", "tif", invalid_lookahead="udm", as_list=as_list
            )

        return stack_path

    def _to_reflectance(
        self,
        band_arr: xr.DataArray,
        band_path: AnyPathType,
        band: BandNames,
        **kwargs,
    ) -> xr.DataArray:
        """
        Converts band to reflectance

        Args:
            band_arr (xr.DataArray): Band array to convert
            band_path (AnyPathType): Band path
            band (BandNames): Band to read
            **kwargs: Other keywords

        Returns:
            xr.DataArray: Band in reflectance
        """
        if self._merged:
            return band_arr
        else:
            # Get MTD XML file
            root, nsmap = self.read_mtd()

            # Open identifier
            refl_coef = None

            for band_mtd in root.iterfind(
                f".//{nsmap[self._nsmap_key]}bandSpecificMetadata"
            ):
                if (
                    int(band_mtd.findtext(f".//{nsmap[self._nsmap_key]}bandNumber"))
                    == self.bands[band].id
                ):
                    refl_coef = float(
                        band_mtd.findtext(
                            f".//{nsmap[self._nsmap_key]}reflectanceCoefficient"
                        )
                    )
                    break

            if refl_coef is None:
                raise InvalidProductError(
                    "Couldn't find any reflectanceCoefficient in the product metadata!"
                )

            # To reflectance
            return band_arr * refl_coef

    def _merge_subdatasets_mtd(self):
        """
        Merge subdataset, when several Planet products avec been ordered together
        Will create a reflectance (if possible) VRT, a UDM/UDM2 VRT and a merged metadata XML file.
        """

        def update_corner_dict(key, lon, lat):
            try:
                lon = lon.iat[0]
            except Exception:
                pass
            try:
                lat = lat.iat[0]
            except Exception:
                pass
            xml.update_txt(
                mtd,
                f"{nsmap[self._nsmap_key]}{key}/{nsmap[self._nsmap_key]}longitude",
                lon,
            )
            xml.update_txt(
                mtd,
                f"{nsmap[self._nsmap_key]}{key}/{nsmap[self._nsmap_key]}latitude",
                lat,
            )

        # Merge datasets
        analytic_vrt_path, analytic_vrt_exists = self._merge_subdatasets()

        # Check if mtd needs an update
        mtd_file, mtd_exists = self._get_out_path(f"{self.condensed_name}_metadata.xml")

        # -- Update VRT
        scales = defaultdict(dict)
        cloud_cover = []
        udp = []
        if not mtd_exists or not analytic_vrt_exists:
            # Get all scales, cloud cloudCoverPercentage, unusableDataPercentage
            for mtd_f in self._get_path("metadata", "xml", as_list=True):
                mtd_filename = path.get_filename(mtd_f)
                subprod_name = mtd_filename.split("_Analytic")[0]
                mtd, nsmap = self._read_mtd_xml(
                    f"{subprod_name}*metadata*xml", f"{subprod_name}.*metadata.*xml"
                )

                # reflectanceCoefficient
                for band_mtd in mtd.iterfind(
                    f".//{nsmap[self._nsmap_key]}bandSpecificMetadata"
                ):
                    band_nb = band_mtd.findtext(f"{nsmap[self._nsmap_key]}bandNumber")
                    refl_coef = band_mtd.findtext(
                        f"{nsmap[self._nsmap_key]}reflectanceCoefficient"
                    )
                    scales[subprod_name][band_nb] = refl_coef

                # cloudCoverPercentage
                cloud_cover.append(
                    float(mtd.findtext(f".//{nsmap['opt']}cloudCoverPercentage"))
                )

                # unusableDataPercentage
                udp.append(
                    float(
                        mtd.findtext(
                            f".//{nsmap[self._nsmap_key]}unusableDataPercentage"
                        )
                    )
                )

        if not analytic_vrt_exists:
            LOGGER.debug("Update raster VRT")
            vrt = etree.parse(analytic_vrt_path).getroot()

            # Remove stats and histograms
            xml.remove(vrt, "Metadata")
            xml.remove(vrt, "Histograms")

            # Convert to Float32
            xml.update_attrib(
                vrt, "VRTRasterBand[@dataType='UInt16']", "dataType", "Float32"
            )  # datatype with d!
            xml.update_attrib(
                vrt, "SourceProperties[@DataType='UInt16']", "DataType", "Float32"
            )  # datatype with D!

            # Scale the VRT
            for el in vrt.iterfind(".//ComplexSource"):
                band_name = path.get_filename(el.findtext("SourceFilename")).split(
                    "_Analytic"
                )[0]
                band_number = el.findtext("SourceBand")

                # Set scaleRatio in VRT
                xml.add(el, "ScaleRatio", scales[band_name][band_number])

            # Write VRT on disk
            xml.write(vrt, analytic_vrt_path)

        # -- Update MTD
        if not mtd_exists:
            LOGGER.debug("Merge metadata")
            mtd, nsmap = self.read_mtd()

            # Remove all reflectance scaling
            xml.remove(mtd, f"{nsmap[self._nsmap_key]}reflectanceCoefficient")

            # Get new size from VRT
            with rasterio.open(str(analytic_vrt_path)) as ds:
                xml.update_txt(mtd, f"{nsmap[self._nsmap_key]}numRows", ds.height)
                xml.update_txt(mtd, f"{nsmap[self._nsmap_key]}numColumns", ds.width)

            # Get new extent from VRT
            extent = rasters.get_extent(analytic_vrt_path)
            extent_wgs84 = extent.to_crs(WGS84)

            # Compute centroid and reproject to WGS84 after
            pos = extent.centroid.to_crs(WGS84).iat[0]
            xml.update_txt(mtd, f"{nsmap['gml']}pos", f"{pos.x} {pos.y}")

            # Get extent coordinates (should be footprint but too long to compute IMHO)
            coordinates_str = " ".join(
                f"{coord[0]},{coord[1]}"
                for coord in extent_wgs84.boundary.iat[0].coords
            )
            xml.update_txt(mtd, f"{nsmap['gml']}coordinates", coordinates_str)

            # Get corners
            bounds_wgs84 = extent_wgs84.bounds
            update_corner_dict("topLeft", bounds_wgs84.maxx, bounds_wgs84.miny)
            update_corner_dict("topRight", bounds_wgs84.maxx, bounds_wgs84.maxy)
            update_corner_dict("bottomRight", bounds_wgs84.minx, bounds_wgs84.maxy)
            update_corner_dict("bottomLeft", bounds_wgs84.minx, bounds_wgs84.miny)

            # Manage cloudCoverPercentage, unusableDataPercentage
            xml.update_txt(
                mtd, f"{nsmap['opt']}cloudCoverPercentage", np.mean(cloud_cover)
            )
            xml.update_txt(
                mtd, f"{nsmap[self._nsmap_key]}unusableDataPercentage", np.mean(udp)
            )

            if self.product_type == PlaProductType.L3A:
                # -- PSOrthoTile

                # identifier: keep the one opened

                # Remove tileId
                xml.remove(mtd, f"{nsmap[self._nsmap_key]}tileId")

                # Round incidenceAngle
                xml.update_txt_fct(
                    mtd,
                    f"{nsmap['eop']}incidenceAngle",
                    lambda x: np.round(float(x), decimals=1),
                )

            elif self.product_type == PlaProductType.L3B:
                # -- PSOrthoScene
                # identifier: replace satellite ID by XX: 20210902_093940_06_245d_3B_AnalyticMS_8b -> 20210902_093940_XX_245d_3B_AnalyticMS_8b
                xml.update_txt_fct(
                    mtd,
                    f"{nsmap['eop']}identifier",
                    lambda x: "_".join(
                        "XX" if i == 2 else z for i, z in enumerate(x.split("_"))
                    ),
                )

                # Remove filename
                xml.remove(mtd, f"{nsmap['eop']}fileName")

                # Round incidenceAngle, illuminationAzimuthAngle, illuminationElevationAngle, azimuthAngle, spaceCraftViewAngle
                xml.update_txt_fct(
                    mtd,
                    f"{nsmap['eop']}incidenceAngle",
                    lambda x: np.round(float(x), decimals=1),
                )
                xml.update_txt_fct(
                    mtd,
                    f"{nsmap['opt']}illuminationAzimuthAngle",
                    lambda x: np.round(float(x), decimals=1),
                )
                xml.update_txt_fct(
                    mtd,
                    f"{nsmap['opt']}illuminationElevationAngle",
                    lambda x: np.round(float(x), decimals=1),
                )
                xml.update_txt_fct(
                    mtd,
                    f"{nsmap[self._nsmap_key]}azimuthAngle",
                    lambda x: np.round(float(x), decimals=1),
                )
                xml.update_txt_fct(
                    mtd,
                    f"{nsmap[self._nsmap_key]}spaceCraftViewAngle",
                    lambda x: np.round(float(x), decimals=1),
                )

            else:
                raise NotImplementedError

            # Write XML on disk
            xml.write(mtd, mtd_file)
